from rdmo.options.providers import Provider
from ..queries import query_sources

class MaRDISearch(Provider):

    search = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi']

        return query_sources(search, query_id, sources)
        
class SoftwareSearch(Provider):

    search = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'SO'
        sources = ['mathalgodb']

        return query_sources(search, query_id, sources)
    



'''Module containing Providers for the Model Documentation'''

from rdmo.options.providers import Provider

from .. helpers import define_setup
from ..queries import query_sources, query_sources_with_user_additions


class ResearchField(Provider):
    '''Research Field Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedResearchFieldWithCreation(Provider):
    '''Research Field Provider (MaRDI Portal / Wikidata),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['field'],
            creation = True
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedResearchFieldWithoutCreation(Provider):
    '''Research Field Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['field'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class ResearchProblem(Provider):
    '''Research Problem Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedResearchProblemWithCreation(Provider):
    '''Research Problem Provider (MaRDI Portal / Wikidata),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['problem'],
            creation = True
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedResearchProblemWithoutCreation(Provider):
    '''Research Problem Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['problem'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class MathematicalModel(Provider):
    '''Mathematical Model Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external sources for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedMathematicalModelWithoutCreation(Provider):
    '''Mathematical Model Provider (MaRDI Portal / Wikidata),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external sources for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['model'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class QuantityOrQuantityKind(Provider):
    '''Quantity [Kind] Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedQuantityWithoutCreation(Provider):
    '''Quantity Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['quantity'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedQuantityKindWithoutCreation(Provider):
    '''Quantity Kind Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['quantity'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedQuantityOrQuantityKindWithCreation(Provider):
    '''Quantity [Kind] Provider (MaRDI Portal / Wikidata),
       User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external source for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['quantity'],
            creation = True
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class MathematicalFormulation(Provider):
    '''Mathematical Formulation Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external sources for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedMathematicalFormulationWithCreation(Provider):
    '''Mathematical Formulation Provider (MaRDI Portal / Wikidata),
       User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external sources for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['formulation'],
            creation = True
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedMathematicalFormulationWithoutCreation(Provider):
    '''Mathematical Formulation Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries external sources for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['formulation'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class Task(Provider):
    '''Task Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh =True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        return query_sources(search)

class RelatedTaskWithCreation(Provider):
    '''Task Provider (MaRDI Portal / Wikidata),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['task'],
            creation = True
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedTaskWithoutCreation(Provider):
    '''Task Provider (MaRDI Portal),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['task'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedModelEntityWithoutCreation(Provider):
    '''Research Field, Research Problem, Mathematical Model,
       Mathematical Formulation, Quantity [Kind], Task Provider 
       (MaRDI Portal / Wikidata), No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['field', 'problem', 'model', 'quantity', 'formulation', 'task'],
            sources = ['mardi']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )



'''General Helper Functions of MaRDMO'''

from typing import Callable, Optional, Any

import re

from rdmo.projects.models import Value
from rdmo.domain.models import Attribute
from rdmo.options.models import Option

def date_format(parts: list[int]) -> str | None:
    """Return a date string as YYYY-MM-DDT00:00:00Z given a list of date parts."""
    if not parts:
        return None
    if len(parts) == 1:
        return f"{int(parts[0]):04d}-00-00T00:00:00Z"
    if len(parts) == 2:
        return f"{int(parts[0]):04d}-{int(parts[1]):02d}-00T00:00:00Z"
    return f"{int(parts[0]):04d}-{int(parts[1]):02d}-{int(parts[2]):02d}T00:00:00Z"

def date_precision(date_str: str) -> int | None:
    """
    Return Wikibase precision based on a date string.
    '+YYYY' -> 9, '+YYYY-MM' -> 10, '+YYYY-MM-DD' -> 11.
    """
    if not date_str:
        return None

    # Remove leading + and split date/time
    date_only = date_str.lstrip('+').split('T')[0]  # 'YYYY-MM-DD' or shorter
    parts = date_only.split('-')

    year = parts[0]
    month = parts[1] if len(parts) > 1 else '00'
    day = parts[2] if len(parts) > 2 else '00'

    if year != '0000':
        if month != '00':
            if day != '00':
                return 11  # day precision
            return 10  # month precision
        return 9  # year precision
    return None


def split_value(
    data: dict,
    key: str,
    transform: Optional[Callable[[str], Any]] = None,
    object_role: Optional[Callable[[Any], bool]] = None,
) -> list:
    """
    Split data[key]['value'] on ' / '. Optionally apply a transform
    to each element and filter the results with `object_role`.
    """
    if key not in data:
        return []

    raw = data.get(key, {}).get('value', '').split(" / ")

    parts = [part for part in raw if part]

    if transform is not None:
        parts = [transform(part) for part in parts]

    if object_role is not None:
        parts = [part for part in parts if object_role(part)]

    return parts

def basic_dict(value):
    '''Basic Information from ID Question as Dict'''
    # Extract Label and Description from ID Question
    label, description, _ = extract_parts(value.text)
    # Return Basic Dict
    return {
        'ID': value.external_id, 
        'Name': label, 
        'Description': description
    }

def basic_list(value):
    '''Basic List with Option URI and Text'''
    # Return Basic List
    return [
        value.option_uri,
        value.text
    ]

def define_setup(query_attributes, creation=False, query_id='', sources=None):
    """Define the setup of particular queries."""
    return {
        'creation': creation,
        'query_attributes': query_attributes,
        'query_id': query_id,
        'sources': sources,
    }

def nested_set(data, path, entry):
    """Walk a sequence of keys, creating dicts as needed, and set the final value."""
    d = data
    for key in path[:-1]:
        d = d.setdefault(key, {})
    d[path[-1]] = entry

def extract_parts(string):
    '''Extract Label, Description and Source from ID Question'''
    # Step 1: Split the string at the last occurrence of ') [' to isolate source
    parts = string.rsplit(') [', 1)
    if len(parts) == 2:
        main_part, c = parts[0].strip(), parts[1].rstrip(']')
    else:
        main_part, c = parts[0].strip(), ""
    # Step 2: Find the last whitespace outside brackets to split
    depth = 0
    split_index = -1
    for i, char in enumerate(main_part):
        if char in ('(', '['):
            depth += 1
        elif char in (')', ']'):
            depth -= 1
        elif char == ' ' and depth == 0:
            split_index = i  # Update split_index to last whitespace outside brackets
    if split_index != -1:
        a = main_part[:split_index].strip()
        b = main_part[split_index+1:].strip().lstrip('(')  # Strip any leading '('
    else:
        a, b = main_part, ""
    return a, b, c

def value_editor(project, uri, info):
    '''Add values to the Questionnaire'''
    attribute_object = Attribute.objects.get(uri=uri)
    # Prepare the defaults dictionary
    defaults = {
        'project': project,
        'attribute': attribute_object,
    }

    if info.get('text') is not None:
        defaults['text'] = info['text']

    if info.get('external_id') is not None:
        defaults['external_id'] = info['external_id']

    if info.get('option') is not None:
        defaults['option'] = Option.objects.get(uri=info['option'])

    # Prepare the fields for update_or_create
    update_fields = {
        'project': project,
        'attribute': attribute_object,
        'defaults': defaults
    }

    if info.get('collection_index') is not None:
        update_fields['collection_index'] = info['collection_index']

    if info.get('set_index') is not None:
        update_fields['set_index'] = info['set_index']

    if info.get('set_prefix') is not None:
        update_fields['set_prefix'] = info['set_prefix']

    # Update or create the value
    obj, created = Value.objects.update_or_create(**update_fields)

    return obj, created

def merge_dicts_with_unique_keys(answers, keys):
    '''Merge Dicts with unique Keys.'''
    merged_dict = {}

    for key in keys:
        for inner_key, value in answers[key].items():
            new_inner_key = f"{inner_key}{key}"
            merged_dict[new_inner_key] = value

    return merged_dict

def check_list(list_var):
    '''Check if List is List'''
    if list_var is None:
        list_var = []
    elif not isinstance(list_var, list):
        list_var = [list_var]
    return list_var

def label_index_map(data, data_type):
    '''Map Label to Index'''
    label_to_index_maps = []
    for to_idx_entry in data_type:
        label_to_index_maps.append(
            {
                data[to_idx_entry][k].get('Name'): idx
                for idx, k in enumerate(data.get(to_idx_entry, {}))
            }
        )
    return label_to_index_maps

def resolve_target(name, id_, entity_enc, label_map):
    """Try to resolve name to index in label_map; fallback to id_."""
    if name in label_map:
        return f"{entity_enc}{label_map[name] + 1}"
    return id_


def build_new_value(from_entry, entity, key, resolved, order):
    """Build the new value depending on relation and order flags."""
    if not entity['relation']:
        return resolved

    relation_value = from_entry.get(entity['relation'], {}).get(
        key, "MISSING RELATION TYPE"
    )

    if order['formulation']:
        return [
            relation_value,
            resolved,
            from_entry.get('formulation_number', {}).get(key),
        ]
    if order['task']:
        return [
            relation_value,
            resolved,
            from_entry.get('task_number', {}).get(key),
        ]
    if not order['formulation'] and not order['task']:
        return [relation_value, resolved]

    return resolved


def entity_relations(data, idx, entity, order):
    """Process Entity Relations."""
    idx['to'] = check_list(idx.get('to'))
    entity['encryption'] = check_list(entity['encryption'])
    label_to_index_maps = label_index_map(data, idx['to'])

    for from_entry in data.get(idx.get('from'), {}).values():
        for key, value in from_entry.get(entity['old_name'], {}).items():
            entity_values = from_entry.setdefault(entity['new_name'], {})

            resolved = None
            for enc_entry, label_map in zip(entity['encryption'], label_to_index_maps):
                resolved = resolve_target(
                    name=value.get("Name"),
                    id_=value.get("ID"),
                    entity_enc=enc_entry,
                    label_map=label_map,
                )
                if resolved != value.get("ID"):
                    break  # match found

            new_value = build_new_value(from_entry, entity, key, resolved, order)

            if key not in entity_values.values() and entity_values.get(key) != new_value:
                entity_values[key] = new_value

def initialize_counter(counter):
    '''Function which initializes counter.'''
    return int(max(counter, default=-1)) + 1

def map_entity(data, idx, entity):
    '''Map Entities'''
    # Ensure idx['to'] and enc are lists
    idx['to'] = check_list(idx['to'])
    entity['encryption'] = check_list(entity['encryption'])

    # Create mappings for all idx['to'] lists
    label_to_index_maps = label_index_map(data, idx['to'])

    # Use Template or Ressource Label
    for from_entry in data.get(idx['from'], {}).values():
        for outer_key, relation in from_entry.get(entity['old_name'], {}).items():
            for inner_key, entity_item in relation.items():
                match_found = False
                # Create Dict Entry
                outer = from_entry.setdefault(entity['new_name'], {})
                inner = outer.setdefault(outer_key, {})
                for enc_entry, label_to_index in zip(entity['encryption'], label_to_index_maps):
                    if entity_item['Name'] in label_to_index:
                        label_idx = label_to_index[entity_item['Name']]
                        match_found = True
                        inner[inner_key] = f"{enc_entry}{label_idx+1}"
                        break

                if not match_found:
                    inner[inner_key] = entity_item['ID']

def process_qualifier(value):
    '''Process Qualifier'''
    # Create Value Dictionary
    value_dict = {}
    # Get splitted Values
    value_splitted = value.split(' <|> ')
    for value_idx, value_text in enumerate(value_splitted):
        # Extract Value ID, Label, and Description
        value_id, value_label, value_description = value_text.split(' | ')
        # Get Value Source
        value_source, _ = value_id.split(':')
        # Add to dict
        value_dict.update({value_idx: {'id': value_id,
                                       'label': value_label,
                                       'description': value_description,
                                       'source': value_source}})
    return value_dict


def reduce_prefix(prefix):
    '''Function that takes a prefix and reduces it if |-appended.'''
    if isinstance(prefix, int):
        prefix_reduced = prefix
    else:
        prefix_reduced = int(prefix.split('|')[0])
    return prefix_reduced

def relation_exists(value, set_prefix_red, info, relation_id=None):
    '''Checks if a value–set (–relation) combination already exists in info.
       If relation_id is provided, also checks relation existence.'''

    # Case: relation check required
    if relation_id and "rels" in info:
        return any(
            (
                (f"{value.label} ({value.description})" in text or vid == value.id)
                and int(sid) == set_prefix_red
                and rel == relation_id
            )
            for vid, sid, rel, text in zip(
                info['value_ids'],
                info['set_prefix_ids'],
                info['rels'],
                info['texts'],
            )
        )

    # Case: only value + set check
    return any(
        (
            (f"{value.label} ({value.description})" in text or vid == value.id)
            and int(sid) == set_prefix_red
        )
        for vid, sid, text in zip(
            info['value_ids'],
            info['set_prefix_ids'],
            info['texts']
        )
    )

def relevant_set_ids(info, set_prefix_red):
    ''''Get relevant Set IDs'''
    relevant_set_ids_list = []
    for set_index, set_prefix in zip(info['set_index_ids'], info['set_prefix_ids']):
        if set_prefix == set_prefix_red:
            relevant_set_ids_list.append(set_index)
    return relevant_set_ids_list

def replace_in_dict(d, target, replacement):
    '''Replace IDs in Dict'''
    if isinstance(d, dict):
        return {k: replace_in_dict(v, target, replacement) for k, v in d.items()}
    if isinstance(d, list):
        return [replace_in_dict(v, target, replacement) for v in d]
    if isinstance(d, str):
        return d.replace(target, replacement)
    return d

def unique_items(data, title = None):
    '''Search unique Items'''
    # Set up Item Dict and track seen Items
    items = {}
    seen_items = set()
    # Add Workflow Item
    if title:
        triple = (
            'not found',
            title,
            data.get('general', {}).get('objective', '')
        )
        items[f'Item{str(0).zfill(10)}'] = {
            'ID': 'not found',
            'Name': title,
            'Description': data.get('general', {}).get('objective', '')
        }
        seen_items.add(triple)
    # Add Workflow Component Items
    def search(subdict):
        '''Search unique Items'''
        if isinstance(subdict, dict) and 'ID' in subdict:
            triple = (
                subdict.get('ID', ''),
                subdict.get('Name', ''),
                subdict.get('Description', ''),
                subdict.get('orcid', ''),
                subdict.get('zbmath', ''),
                subdict.get('issn', '')
            )
            if triple not in seen_items:
                item_key = f'Item{str(len(items)).zfill(10)}'  # Create unique key
                items[item_key] = {
                    'ID': triple[0],
                    'Name': triple[1],
                    'Description': triple[2],
                    'orcid': triple[3],
                    'zbmath': triple[4],
                    'issn': triple[5]
                }
                seen_items.add(triple)
        if isinstance(subdict, dict):
            for value in subdict.values():
                if isinstance(value, dict):
                    search(value)
    search(data)
    return items

def inline_mathml(data):
    '''Process MathML to be inline'''
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, str):
                if '<math' in value:
                    data[key] = clean_mathml(value)
            elif isinstance(value, (dict, list)):
                inline_mathml(value)
    elif isinstance(data, list):
        for item in data:
            inline_mathml(item)

def clean_mathml(mathml_str):
    '''Clean MathL'''
    def clean_tag(match):
        '''Clean Tag'''
        tag = match.group(1)

        # Keep xmlns on <math> tag
        if tag.startswith('math'):
            xmlns_match = re.search(r'xmlns="[^"]+"', tag)
            if xmlns_match:
                return f"<math {xmlns_match.group(0)}>"
            return "<math>"

        # Just keep the tag name, strip attributes
        tagname_match = re.match(r'^/?\w+', tag)
        if tagname_match:
            return f"<{tagname_match.group(0)}>"
        return f"<{tag}>"

    # Apply substitution on all opening tags
    cleaned = re.sub(r'<([^>\s]+(?:\s[^>]*)?)>', clean_tag, mathml_str)
    return cleaned

def process_question_dict(project, questions, get_answer):
    """Iterate through nested questions dict and extract answers using provided answer function."""
    answers = {}

    for group in questions.values():
        for sub_key, config in group.items():
            if sub_key == "uri":
                continue  # Skip the group-level URI

            if not isinstance(config, dict) or "uri" not in config:
                continue  # Skip invalid or metadata-only entries

            # Fill in optional/default values
            config = {
                "key1": config.get("key1"),
                "key2": config.get("key2"),
                "key3": config.get("key3"),
                "uri": config["uri"],
                "set_prefix": config.get("set_prefix", False),
                "set_index": config.get("set_index", False),
                "collection_index": config.get("collection_index", False),
                "external_id": config.get("external_id", False),
                "option_text": config.get("option_text", False),
            }

            # Call the injected function
            answers = get_answer(
                project,
                answers,
                config
            )

    return answers

def compare_items(old, new):
    """Compare Items before and after Export, to list newly created Items"""
    ids = {}
    for key, value in old.items():
        if key.startswith('Item') and not value['id']:
            ids.update({new[key]['payload']['item']['labels']['en']: new[key]['id']})
    return ids



'''BASE URI and Endpoints required by MaRDMO'''

#RDMO BASE URI
BASE_URI = 'https://rdmo.mardi4nfdi.de/terms/'

# Endpoints
endpoint = {
    'mardi': {
        'api': 'https://portal.mardi4nfdi.de/w/api.php',
        'sparql': 'https://query.portal.mardi4nfdi.de/sparql',
        'uri': 'https://portal.mardi4nfdi.de'
              },
    'wikidata': {
        'api': 'https://www.wikidata.org/w/api.php',
        'sparql': 'https://query-main.wikidata.org/sparql',
        'sparql-scholarly': 'https://query-scholarly.wikidata.org/sparql',
        'uri': 'https://www.wikidata.org/wiki/Item:'
                },
    'mathalgodb': {
        'sparql': 'https://sparql.cordi2025.m1.mardi.ovh/mathalgodb/query',
        'update': 'https://sparql.cordi2025.m1.mardi.ovh/mathalgodb/update',
        'uri': 'https://cordi2025.m1.mardi.ovh/'
                  },
    'crossref': {
        'api': 'https://api.crossref.org/works/'
                 },
    'datacite': {
        'api': 'https://api.datacite.org/dois/'
                 },
    'doi': {
        'api': 'https://citation.doi.org/metadata?doi='
            },
    'zbmath': {
        'api': ('https://api.zbmath.org/v1/document/_structured_search'
                '?page=0&results_per_page=100&DOI=')
              },
    'orcid': {
        'api': 'https://pub.orcid.org/v3.0'
             }
    }



from dataclasses import dataclass, field
from typing import List, Optional

from ..adders import add_reference_order
from ..models import Relatant

@dataclass
class Benchmark:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    reference: Optional[List] = field(default_factory=list)
    publications: Optional[List[Relatant]] = field(default_factory=list)
    

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Benchmark':

        data = raw_data[0]
        order = add_reference_order('benchmark')

        return cls(
            id = None,
            label = None,
            description = None,
            reference = {order[key][0]: [order[key][1], value] for part in data.get('reference', {}).get('value', '').split(' | ') if (key := part.split(':')[0]) in order and (value := part.split(':')[1])} | ({order['url'][0]: [order['url'][1], url]} if (url := next((part for part in data.get('reference', {}).get('value', '').split(' | ') if part.startswith('https://')), None)) else {}),
            publications = [Relatant.from_query(publication) for publication in data.get('publication', {}).get('value', '').split(" / ") if publication] if 'publication' in data else []
        )
        
@dataclass
class Software:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    reference: Optional[List] = field(default_factory=list)
    tests: Optional[List[Relatant]] = field(default_factory=list)
    publications: Optional[List[Relatant]] = field(default_factory=list)
    

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Software':

        data = raw_data[0]
        order = add_reference_order('software')

        return cls(
            id = None,
            label = None,
            description = None,
            reference = {order[key][0]: [order[key][1], value] for part in data.get('reference', {}).get('value', '').split(' | ') if (key := part.split(':')[0]) in order and (value := part.split(':')[1])} | ({order['url'][0]: [order['url'][1], url]} if (url := next((part for part in data.get('reference', {}).get('value', '').split(' | ') if part.startswith('https://')), None)) else {}),
            tests =  [Relatant.from_query(benchmark) for benchmark in data.get('tests', {}).get('value', '').split(" / ") if benchmark] if 'tests' in data else [], 
            publications = [Relatant.from_query(publication) for publication in data.get('publication', {}).get('value', '').split(" / ") if publication] if 'publication' in data else []
        )
    
@dataclass
class Problem:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    specializes: Optional[List[Relatant]] = field(default_factory=list)
    specializedBy: Optional[List[Relatant]] = field(default_factory=list)
    instantiates: Optional[List[Relatant]] = field(default_factory=list)
    

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Problem':

        data = raw_data[0]

        return cls(
            id = None,
            label = None,
            description = None,
            specializes = [Relatant.from_query(specializes) for specializes in data.get('specializes', {}).get('value', '').split(" / ") if specializes] if 'specializes' in data else [], 
            specializedBy = [Relatant.from_query(specializedBy) for specializedBy in data.get('specializedBy', {}).get('value', '').split(" / ") if specializedBy] if 'specializedBy' in data else [], 
            instantiates =  [Relatant.from_query(benchmark) for benchmark in data.get('instantiates', {}).get('value', '').split(" / ") if benchmark] if 'instantiates' in data else [], 
        )
    
@dataclass
class Algorithm:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    componentOf: Optional[List[Relatant]] = field(default_factory=list)
    hasComponent: Optional[List[Relatant]] = field(default_factory=list)
    subclassOf: Optional[List[Relatant]] = field(default_factory=list)
    hasSubclass: Optional[List[Relatant]] = field(default_factory=list)
    relatedTo: Optional[List[Relatant]] = field(default_factory=list)
    solves: Optional[List[Relatant]] = field(default_factory=list)
    implementedBy: Optional[List[Relatant]] = field(default_factory=list)
    publications: Optional[List[Relatant]] = field(default_factory=list)
    

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Algorithm':

        data = raw_data[0]

        return cls(
            id = None,
            label = None,
            description = None,
            componentOf = [Relatant.from_query(componentOf) for componentOf in data.get('componentOf', {}).get('value', '').split(" / ") if componentOf] if 'componentOf' in data else [], 
            hasComponent = [Relatant.from_query(hasComponent) for hasComponent in data.get('hasComponent', {}).get('value', '').split(" / ") if hasComponent] if 'hasComponent' in data else [], 
            subclassOf = [Relatant.from_query(subclassOf) for subclassOf in data.get('subclassOf', {}).get('value', '').split(" / ") if subclassOf] if 'subclassOf' in data else [], 
            hasSubclass = [Relatant.from_query(hasSubclass) for hasSubclass in data.get('hasSubclass', {}).get('value', '').split(" / ") if hasSubclass] if 'hasSubclass' in data else [], 
            relatedTo = [Relatant.from_query(relatedTo) for relatedTo in data.get('relatedTo', {}).get('value', '').split(" / ") if relatedTo] if 'relatedTo' in data else [], 
            solves =  [Relatant.from_query(problem) for problem in data.get('solves', {}).get('value', '').split(" / ") if problem] if 'solves' in data else [], 
            implementedBy =  [Relatant.from_query(software) for software in data.get('implementedBy', {}).get('value', '').split(" / ") if software] if 'implementedBy' in data else [], 
            publications = [Relatant.from_query(publication) for publication in data.get('publication', {}).get('value', '').split(" / ") if publication] if 'publication' in data else []
        )


'''Module containing Constants for the Algorithm Documentation'''

from ..config import BASE_URI
from ..getters import get_questions

# Dictionary with list of property names
PROPS = {
    'A2P': ['solves'],
    'A2S': ['implementedBy'],
    'P2B': ['instantiates'],
    'S2B': ['tests'],
    'Algorithm': ['hasComponent', 'componentOf', 'hasSubclass', 'subclassOf', 'relatedTo'],
    'Problem': ['specializes', 'specializedBy']
}

# URI PREFIX Map
def get_uri_prefix_map():
    '''URI Prefixes for the Algorithm Documentation'''
    questions = get_questions('algorithm')
    uri_prefix_map = {
        f'{BASE_URI}{questions["Problem"]["BRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Benchmark"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Benchmark"]["ID"]["uri"]}',
            "prefix": "B"
        },
        f'{BASE_URI}{questions["Software"]["BRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Benchmark"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Benchmark"]["ID"]["uri"]}',
            "prefix": "B"
        },
        f'{BASE_URI}{questions["Algorithm"]["PRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Problem"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Problem"]["ID"]["uri"]}',
            "prefix": "AP"
        },
        f'{BASE_URI}{questions["Algorithm"]["SRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Software"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Software"]["ID"]["uri"]}',
            "prefix": "S"
        }
    }
    return uri_prefix_map

class_prefix_map = {
        'algorithm': 'al',
        'problem': 'pr',
        'benchmark': 'bm',
        'software': 'sw',
        'publication': 'pb',
    }

# Parameter for Entity relations
PREVIEW_RELATIONS = [
    # fromIDX, toIDX, relationOld, entityOld, entityNew, enc
    ('algorithm', 'problem', 'A2P', 'PRelatant', 'RelationP', 'AP'),
    ('algorithm', 'software', 'A2S', 'SRelatant', 'RelationS', 'S'),
    ('algorithm', 'algorithm', 'IntraClassRelation', 'IntraClassElement', 'RelationA', 'A'),
    ('problem', 'benchmark', 'P2B', 'BRelatant', 'RelationB', 'B'),
    ('problem', 'problem', 'IntraClassRelation', 'IntraClassElement', 'RelationP', 'AP'),
    ('software', 'benchmark', 'S2B', 'BRelatant', 'RelationB', 'B'),
    ('publication', 'algorithm', 'P2A', 'ARelatant', 'RelationA', 'A'),
    ('publication', 'benchmark', 'P2B', 'BRelatant', 'RelationB', 'B'),
    ('publication', 'software', 'P2S', 'SRelatant', 'RelationS', 'S')
]



'''Module containing Handlers for the Publication Documentation'''

from django.dispatch import receiver
from django.db.models.signals import post_delete

from rdmo.projects.models import Value
from rdmo.options.models import Option
from rdmo.domain.models import Attribute

from ..config import BASE_URI, endpoint
from ..getters import get_items, get_mathmoddb, get_mathalgodb, get_properties, get_questions
from ..helpers import value_editor
from ..queries import query_sparql
from ..adders import add_basics, add_references, add_relations_flexible

from .constants import (
    INDEX_COUNTERS,
    PROPS,
    RELATANT_URIS,
    RELATION_URIS,
    PUBLICATIONS,
    LANGUAGES,
    JOURNALS,
    AUTHORS
)
from .utils import generate_label
from .sparql import queryPublication
from .models import Publication

class Information:
    '''Class containing functions, querying external sources for specific
       entities and integrating the related metadata into the questionnaire.'''

    def __init__(self):
        '''Load shared data once'''
        self.questions = get_questions('publication')
        self.mathalgodb = get_mathalgodb()
        self.mathmoddb = get_mathmoddb()
        self.base = BASE_URI

    def citation(self, instance):
        '''Citation Information'''

        # Publication specific Questions.
        publication = self.questions["Publication"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Get Source and ID of selected Publication
        source, identifier = instance.external_id.split(':')

        # Empty Data as Fallback
        data = {}

        # If Publication from MathAlgoDB...
        if source == 'mathalgodb':
            #...query the MathAlgoDB,...
            query = queryPublication['PublicationMathAlgoDB'].format(
                identifier
            )
            results = query_sparql(
                query,
                endpoint['mathalgodb']['sparql']
            )

            if not results:
                return

            #...structure the data...
            data = Publication.from_query(results)

            #...and add the Information to the Questionnaire.
            add_basics(
                project = instance.project,
                text = instance.text,
                questions = self.questions,
                item_type = 'Publication',
                index = (instance.set_index, None)
            )

            add_references(
                project = instance.project,
                data = data,
                uri = f'{BASE_URI}{publication["Reference"]["uri"]}',
                set_index = instance.set_index
            )

        # If Publication from MaRDI Portal...
        elif source == 'mardi':
            #...query the MaRDI Portal,...
            query = queryPublication['MaRDI']['QID_FULL'].format(
                identifier,
                **get_items(),
                **get_properties()
            )
            results = query_sparql(
                query,
                endpoint['mardi']['sparql']
            )

            if not results:
                return

            #...structure the data...
            data = Publication.from_query(results)

            #...and add the Information to the Questionnaire.
            if str(instance.project.catalog).endswith('mardmo-algorithm-catalog'):
                value_editor(
                    project = instance.project,
                    uri = f'{BASE_URI}{publication["Name"]["uri"]}',
                    info = {
                        'text': generate_label(data), 
                        'set_index': instance.set_index
                    }
                )
            else:
                value_editor(
                    project = instance.project,
                    uri = f'{BASE_URI}{publication["Name"]["uri"]}',
                    info = {
                        'text': data.label, 
                        'set_index': instance.set_index
                    }
                )

            value_editor(
                project = instance.project,
                uri = f'{BASE_URI}{publication["Description"]["uri"]}',
                info = {
                    'text': data.description, 
                    'set_index': instance.set_index
                }
            )

            add_references(
                project = instance.project,
                data = data,
                uri = f'{BASE_URI}{publication["Reference"]["uri"]}',
                set_index = instance.set_index
            )

        # If Publication from Wikidata...
        elif source == 'wikidata':
            #...query Wikidata,...
            query = queryPublication['Wikidata']['QID_FULL'].format(
                identifier
            )
            results = query_sparql(
                query,
                endpoint['wikidata']['sparql-scholarly']
            )

            if not results:
                return

            #...structure the data...
            data = Publication.from_query(results)

            #and add the Information to the Questionnaire.
            if str(instance.project.catalog).endswith('mardmo-algorithm-catalog'):
                value_editor(
                    project = instance.project,
                    uri = f'{BASE_URI}{publication["Name"]["uri"]}',
                    info = {
                        'text': generate_label(data), 
                        'set_index': instance.set_index
                    }
                )
            else:
                value_editor(
                    project = instance.project,
                    uri = f'{BASE_URI}{publication["Name"]["uri"]}',
                    info = {
                        'text': data.label, 
                        'set_index': instance.set_index
                    }
                )

            value_editor(
                project = instance.project,
                uri = f'{BASE_URI}{publication["Description"]["uri"]}',
                info = {
                    'text': data.description, 
                    'set_index': instance.set_index
                }
            )

            add_references(
                project = instance.project,
                data = data,
                uri = f'{BASE_URI}{publication["Reference"]["uri"]}',
                set_index = instance.set_index
            )

        if data:

            # For Models add Relations
            if str(instance.project.catalog).endswith('mardmo-model-catalog'):
                if source == 'mardi':
                    add_relations_flexible(
                        project = instance.project,
                        data = data,
                        props = {
                            'keys': PROPS['P2E'],
                            'mapping': self.mathmoddb,
                        },
                        index = {
                            'set_prefix': instance.set_index
                        },
                        statement = {
                            'relation': f'{BASE_URI}{publication["P2E"]["uri"]}',
                            'relatant': f'{BASE_URI}{publication["EntityRelatant"]["uri"]}',
                        },
                    )

            # For Algorithms add Relations
            if str(instance.project.catalog).endswith('mardmo-algorithm-catalog'):
                if source == 'mathalgodb':
                    add_relations_flexible(
                        project = instance.project,
                        data = data,
                        props = {
                            'keys': PROPS['P2A'],
                            'mapping': self.mathalgodb,
                        },
                        index = {
                            'set_prefix': instance.set_index
                        },
                        statement = {
                            'relation': f'{BASE_URI}{publication["P2A"]["uri"]}',
                            'relatant': f'{BASE_URI}{publication["ARelatant"]["uri"]}',
                        },
                    )

                    for prop in PROPS['P2BS']:
                        for value in getattr(data, prop):
                            value_editor(
                                project = instance.project,
                                uri = f'{BASE_URI}{publication[RELATION_URIS[value.item_class]]["uri"]}',
                                info = {
                                    'option': Option.objects.get(uri = self.mathalgodb[prop]),
                                    'set_index': INDEX_COUNTERS[value.item_class],
                                    'set_prefix': instance.set_index
                                }
                            )
                            value_editor(
                                project = instance.project,
                                uri = f'{BASE_URI}{publication[RELATANT_URIS[value.item_class]]["uri"]}',
                                info = {
                                    'text': f"{value.label} ({value.description}) [mathalgodb]",
                                    'external_id': value.id,
                                    'set_index': INDEX_COUNTERS[value.item_class],
                                    'set_prefix': instance.set_index
                                }
                            )
                            INDEX_COUNTERS[value.item_class] += 1
        return

@receiver(post_delete, sender=Value)
def publication_delete(sender, **kwargs):
    '''Handler to delete hidden Publication Information
       upon deletion of Set.
    '''
    instance = kwargs.get("instance", None)
    # Get Questions of Publication Section
    questions = get_questions('publication')
    # Check if Publication is concerned
    if instance and instance.attribute.uri == f'{BASE_URI}{questions["Publication"]["uri"]}':
        # Get Project, Snapshot and Index of Deleted Set
        set_index = instance.set_index
        project = instance.project
        snapshot = instance.snapshot
        # Loop through "hidden" Data and delete it
        for key in PUBLICATIONS | LANGUAGES | JOURNALS | AUTHORS:
            Value.objects.filter(
                attribute_id = Attribute.objects.get(
                    uri = f'{BASE_URI}{questions["Publication"][key]["uri"]}'
                ),
                set_index = set_index,
                project = project,
                snapshot = snapshot
            ).delete()



'''Module containing general Handlers for MaRDMO'''

from .adders import add_entities, add_new_entities
from .helpers import extract_parts
from .models import Relatant

from .algorithm.constants import get_uri_prefix_map as get_uri_prefix_map_algorithm
from .model.constants import get_uri_prefix_map as get_uri_prefix_map_model
from .workflow.constants import get_uri_prefix_map as get_uri_prefix_map_workflow

class Information:
    '''Class containing functions, querying external sources for specific
       entities and integrating the related metadata into the questionnaire.'''

    def __init__(self):
        # Load shared data once
        return

    def relation(self, instance):
        '''Relation Information'''

        # Get appropriate config map
        if str(instance.project.catalog).endswith("mardmo-model-catalog"):
            config_map = get_uri_prefix_map_model()
        elif str(instance.project.catalog).endswith("mardmo-interdisciplinary-workflow-catalog"):
            config_map = get_uri_prefix_map_workflow()
        elif str(instance.project.catalog).endswith("mardmo-algorithm-catalog"):
            config_map = get_uri_prefix_map_algorithm()
        else:
            return

        # Stop if no Text
        if not instance.text:
            return

        # Get item, config and data information
        label, description, source = extract_parts(instance.text)
        config = config_map[instance.attribute.uri]
        datas = [Relatant.from_triple(instance.external_id, label, description)]

        # Add items from specific source
        if source in ('mardi', 'mathalgodb', 'wikidata'):
            add_entities(
                project=instance.project,
                question_set=config["question_set"],
                datas=datas,
                source=source,
                prefix=config["prefix"]
            )

        # Add items from user
        elif source == 'user':
            add_new_entities(
                project=instance.project,
                question_set=config["question_set"],
                datas=datas,
                prefix=config["prefix"]
            )



#Queries to MaRDI KG and Wikidata for Publication Handler

queryPublication = {
    
    'MaRDI': {

        'DOI': '''SELECT ?id ?label ?description         
                        
                 WHERE {{?idraw wdt:{DOI} "{0}".
                        BIND(CONCAT("mardi:", STRAFTER(STR(?idraw), STR(wd:))) AS ?id)
                        
                        OPTIONAL {{?idraw rdfs:label ?labelraw.
                                  FILTER (lang(?labelraw) = 'en')}}
                        BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)
                        
                        OPTIONAL {{?idraw schema:description ?descriptionraw.
                                  FILTER (lang(?descriptionraw) = 'en')}}
                        BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)
                       }}''',
    
       'QID_FULL': '''SELECT ?label ?description ?doi         
                            (GROUP_CONCAT(DISTINCT(?authorInfo); separator=" | ") AS ?authorInfos) 
                            (GROUP_CONCAT(DISTINCT CONCAT(?inv, " | ", ?invl, " | ", ?invd); separator=" / ") AS ?invents)    
                            (GROUP_CONCAT(DISTINCT CONCAT(?doc, " | ", ?docl, " | ", ?docd); separator=" / ") AS ?documents)
                            (GROUP_CONCAT(DISTINCT CONCAT(?stu, " | ", ?stul, " | ", ?stud); separator=" / ") AS ?studies)
                            (GROUP_CONCAT(DISTINCT CONCAT(?rev, " | ", ?revl, " | ", ?revd); separator=" / ") AS ?surveys)
                            (GROUP_CONCAT(DISTINCT CONCAT(?use, " | ", ?usel, " | ", ?used); separator=" / ") AS ?uses)    
                            ?entrytypelabel ?journalInfo ?languagelabel                               
                            ?title ?date ?volume ?issue ?page          

                 WHERE {{ VALUES ?publication {{ wd:{0} }}

                        OPTIONAL {{
                                       ?invraw p:{described by source} ?statement1.
                                       ?statement1 ps:{described by source} ?publication.

                                       BIND(CONCAT("mardi:", STRAFTER(STR(?invraw), STR(wd:))) AS ?inv)
                                       
                                       OPTIONAL {{
                                                  ?invraw rdfs:label ?invlraw.
                                                  FILTER (lang(?invlraw) = 'en')
                                               }}
                                      BIND(COALESCE(?invlraw, "No Label Provided!") AS ?invl)
                                      
                                      OPTIONAL {{
                                                 ?invraw schema:description ?invdraw
                                                 FILTER (lang(?invdraw) = 'en')
                                               }}
                                      BIND(COALESCE(?invdraw, "No Description Provided!") AS ?invd)
                                      
                                      ?statement1 pq:{object has role} wd:{invention}.
                                   }}

                        OPTIONAL {{
                                       ?docraw p:{described by source} ?statement2.
                                       ?statement2 ps:{described by source} ?publication.

                                       BIND(CONCAT("mardi:", STRAFTER(STR(?docraw), STR(wd:))) AS ?doc)
                                       
                                       OPTIONAL {{
                                                  ?docraw rdfs:label ?doclraw.
                                                  FILTER (lang(?doclraw) = 'en')
                                               }}
                                      BIND(COALESCE(?doclraw, "No Label Provided!") AS ?docl)
                                      
                                      OPTIONAL {{
                                                 ?docraw schema:description ?docdraw
                                                 FILTER (lang(?docdraw) = 'en')
                                               }}
                                      BIND(COALESCE(?docdraw, "No Description Provided!") AS ?docd)
                                      
                                      ?statement2 pq:{object has role} wd:{documentation}.
                                   }}

                        OPTIONAL {{
                                       ?sturaw p:{described by source} ?statement3.
                                       ?statement3 ps:{described by source} ?publication.

                                       BIND(CONCAT("mardi:", STRAFTER(STR(?sturaw), STR(wd:))) AS ?stu)
                                       
                                       OPTIONAL {{
                                                  ?sturaw rdfs:label ?stulraw.
                                                  FILTER (lang(?stulraw) = 'en')
                                               }}
                                      BIND(COALESCE(?stulraw, "No Label Provided!") AS ?stul)
                                      
                                      OPTIONAL {{
                                                 ?sturaw schema:description ?studraw
                                                 FILTER (lang(?studraw) = 'en')
                                               }}
                                      BIND(COALESCE(?studraw, "No Description Provided!") AS ?stud)
                                      
                                      ?statement3 pq:{object has role} wd:{study}.
                                   }}

                        OPTIONAL {{
                                       ?revraw p:{described by source} ?statement3.
                                       ?statement3 ps:{described by source} ?publication.

                                       BIND(CONCAT("mardi:", STRAFTER(STR(?revraw), STR(wd:))) AS ?rev)
                                       
                                       OPTIONAL {{
                                                  ?revraw rdfs:label ?revlraw.
                                                  FILTER (lang(?revlraw) = 'en')
                                               }}
                                      BIND(COALESCE(?revlraw, "No Label Provided!") AS ?revl)
                                      
                                      OPTIONAL {{
                                                 ?revraw schema:description ?revdraw
                                                 FILTER (lang(?revdraw) = 'en')
                                               }}
                                      BIND(COALESCE(?revdraw, "No Description Provided!") AS ?revd)
                                      
                                      ?statement3 pq:{object has role} wd:{review}.
                                   }}

                        OPTIONAL {{
                                       ?useraw p:{described by source} ?statement4.
                                       ?statement4 ps:{described by source} ?publication.

                                       BIND(CONCAT("mardi:", STRAFTER(STR(?useraw), STR(wd:))) AS ?use)
                                       
                                       OPTIONAL {{
                                                  ?useraw rdfs:label ?uselraw.
                                                  FILTER (lang(?uselraw) = 'en')
                                               }}
                                      BIND(COALESCE(?uselraw, "No Label Provided!") AS ?usel)
                                      
                                      OPTIONAL {{
                                                 ?useraw schema:description ?usedraw
                                                 FILTER (lang(?usedraw) = 'en')
                                               }}
                                      BIND(COALESCE(?usedraw, "No Description Provided!") AS ?used)
                                      
                                      ?statement4 pq:{object has role} wd:{use}.
                                   }}
                 
                        OPTIONAL {{ ?publication wdt:{DOI} ?doi. }}
                        
                        OPTIONAL {{?publication rdfs:label ?labelraw.
                                  FILTER (lang(?labelraw) = 'en')}}
                        BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)
                        
                        OPTIONAL {{?publication schema:description ?descriptionraw.
                                  FILTER (lang(?descriptionraw) = 'en')}}
                        BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)
                        
                        OPTIONAL {{?publication (wdt:{author} | wdt:{author name string}) ?authorraw.
                                  BIND(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("mardi:", STRAFTER(STR(?authorraw), STR(wd:))), "") AS ?author)
     
                                  OPTIONAL {{?authorraw rdfs:label ?authorlabelraw.
                                            FILTER (lang(?authorlabelraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorlabelraw, ?authorraw), "") AS ?authorlabel)
   
                                  OPTIONAL {{?authorraw schema:description ?authordescriptionraw.
                                            FILTER (lang(?authordescriptionraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authordescriptionraw, ""), "") AS ?authordescription)
   
                                  OPTIONAL {{?authorraw wdt:{ORCID iD} ?authororcidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authororcidraw, ""), "") AS ?authororcid)
                          
                                  OPTIONAL {{?authorraw wdt:{Wikidata QID} ?authorwikidataidraw}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorwikidataidraw), STR(wd:))), ""), "") AS ?authorwikidataid)
   
                                  OPTIONAL {{?authorraw wdt:{zbMATH author ID} ?authorzbmathidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorzbmathidraw, ""), "") AS ?authorzbmathid)
     
                                  BIND(CONCAT(?author, " <|> ", ?authorlabel, " <|> ", ?authordescription, " <|> ", ?authororcid, " <|> ", ?authorzbmathid, " <|> ", ?authorwikidataid) AS ?authorInfo)}}
   
                        OPTIONAL {{?publication wdt:{instance of} ?entrytyperaw.
                                  
                                   OPTIONAL {{?entrytyperaw rdfs:label ?entrytypelabelraw.
                                            FILTER (lang(?entrytypelabelraw) = 'en')}}
                                   BIND(COALESCE(?entrytypelabelraw, "No Label Provided!") As ?entrytypelabel)
                                  }}
                        
                        OPTIONAL {{?publication wdt:{published in} ?journalraw.
                                  BIND(CONCAT("mardi:", STRAFTER(STR(?journalraw), STR(wd:))) AS ?journal)
                                  
                                  OPTIONAL {{?journalraw rdfs:label ?journallabelraw.
                                            FILTER (lang(?journallabelraw) = 'en')}}
                                  BIND(COALESCE(?journallabelraw, "No Label Provided!") As ?journallabel)
                                  
                                  OPTIONAL {{?journalraw schema:description ?journaldescriptionraw.
                                            FILTER (lang(?journaldescriptionraw) = 'en')}}
                                  BIND(COALESCE(?journaldescriptionraw, "No Description Provided!") As ?journaldescription)
                                  
                                  BIND(concat(?journal, " <|> ", ?journallabel, " <|> ", ?journaldescription) AS ?journalInfo)}}
                        
                        OPTIONAL {{?publication wdt:{language of work or name} ?languageraw.
                                   
                                   OPTIONAL {{?languageraw rdfs:label ?languagelabelraw.
                                              FILTER (lang(?languagelabelraw) = 'en')}}
                                   BIND(COALESCE(?languagelabelraw, "No Label Provided!") As ?languagelabel)}}
                        
                        OPTIONAL {{?publication wdt:{title} ?titleraw.
                                   BIND(COALESCE(?titleraw, "No Title Provided!") As ?title)}}
                        
                        OPTIONAL {{?publication wdt:{publication date} ?dateraw.
                                   BIND(COALESCE(?dateraw, "No Publication Date Provided!") As ?date)}}
                        
                        OPTIONAL {{?publication wdt:{volume} ?volumeraw.
                                   BIND(COALESCE(?volumeraw, "No Volume Provided!") As ?volume)}}
                        
                        OPTIONAL {{?publication wdt:{issue} ?issueraw.
                                   BIND(COALESCE(?issueraw, "No Issue Provided!") As ?issue)}}
                        
                        OPTIONAL {{?publication wdt:{page(s)} ?pageraw.
                                   BIND(COALESCE(?pageraw, "No Pages Provided!") As ?page)}}
                        }}
   
                 GROUP BY ?doi ?label ?description ?entrytypelabel ?journalInfo ?languagelabel ?title ?date ?volume ?issue ?page ''',
    
       'DOI_FULL': '''SELECT ?id ?label ?description         
                        (GROUP_CONCAT(DISTINCT(?authorInfo); separator=" | ") AS ?authorInfos)                                                        
                        ?entrytypelabel ?journalInfo ?languagelabel                               
                        ?title ?date ?volume ?issue ?page          

                 WHERE {{?idraw wdt:{DOI} "{0}".
                        BIND(CONCAT("mardi:", STRAFTER(STR(?idraw), STR(wd:))) AS ?id)
                        
                        OPTIONAL {{?idraw rdfs:label ?labelraw.
                                  FILTER (lang(?labelraw) = 'en')}}
                        BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)
                        
                        OPTIONAL {{?idraw schema:description ?descriptionraw.
                                  FILTER (lang(?descriptionraw) = 'en')}}
                        BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)
                        
                        OPTIONAL {{?idraw (wdt:{author} | wdt:{author name string}) ?authorraw.
                                  BIND(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("mardi:", STRAFTER(STR(?authorraw), STR(wd:))), "") AS ?author)
     
                                  OPTIONAL {{?authorraw rdfs:label ?authorlabelraw.
                                            FILTER (lang(?authorlabelraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorlabelraw, ?authorraw), "") AS ?authorlabel)
   
                                  OPTIONAL {{?authorraw schema:description ?authordescriptionraw.
                                            FILTER (lang(?authordescriptionraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authordescriptionraw, ""), "") AS ?authordescription)
   
                                  OPTIONAL {{?authorraw wdt:{ORCID iD} ?authororcidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authororcidraw, ""), "") AS ?authororcid)
                          
                                  OPTIONAL {{?authorraw wdt:{Wikidata QID} ?authorwikidataidraw}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorwikidataidraw), STR(wd:))), ""), "") AS ?authorwikidataid)
   
                                  OPTIONAL {{?authorraw wdt:{zbMATH author ID} ?authorzbmathidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorzbmathidraw, ""), "") AS ?authorzbmathid)
     
                                  BIND(CONCAT(?author, " <|> ", ?authorlabel, " <|> ", ?authordescription, " <|> ", ?authororcid, " <|> ", ?authorzbmathid, " <|> ", ?authorwikidataid) AS ?authorInfo)}}
   
                        OPTIONAL {{?idraw wdt:{instance of} ?entrytyperaw.
                                  
                                   OPTIONAL {{?entrytyperaw rdfs:label ?entrytypelabelraw.
                                            FILTER (lang(?entrytypelabelraw) = 'en')}}
                                   BIND(COALESCE(?entrytypelabelraw, "No Label Provided!") As ?entrytypelabel)
                                  }}
                        
                        OPTIONAL {{?idraw wdt:{published in} ?journalraw.
                                  BIND(CONCAT("mardi:", STRAFTER(STR(?journalraw), STR(wd:))) AS ?journal)
                                  
                                  OPTIONAL {{?journalraw rdfs:label ?journallabelraw.
                                            FILTER (lang(?journallabelraw) = 'en')}}
                                  BIND(COALESCE(?journallabelraw, "No Label Provided!") As ?journallabel)
                                  
                                  OPTIONAL {{?journalraw schema:description ?journaldescriptionraw.
                                            FILTER (lang(?journaldescriptionraw) = 'en')}}
                                  BIND(COALESCE(?journaldescriptionraw, "No Description Provided!") As ?journaldescription)
                                  
                                  BIND(concat(?journal, " <|> ", ?journallabel, " <|> ", ?journaldescription) AS ?journalInfo)}}
                        
                        OPTIONAL {{?idraw wdt:{language of work or name} ?languageraw.
                                   
                                   OPTIONAL {{?languageraw rdfs:label ?languagelabelraw.
                                              FILTER (lang(?languagelabelraw) = 'en')}}
                                   BIND(COALESCE(?languagelabelraw, "No Label Provided!") As ?languagelabel)}}
                        
                        OPTIONAL {{?idraw wdt:{title} ?titleraw.
                                   BIND(COALESCE(?titleraw, "No Title Provided!") As ?title)}}
                        
                        OPTIONAL {{?idraw wdt:{publication date} ?dateraw.
                                   BIND(COALESCE(?dateraw, "No Publication Date Provided!") As ?date)}}
                        
                        OPTIONAL {{?idraw wdt:{volume} ?volumeraw.
                                   BIND(COALESCE(?volumeraw, "No Volume Provided!") As ?volume)}}
                        
                        OPTIONAL {{?idraw wdt:{issue} ?issueraw.
                                   BIND(COALESCE(?issueraw, "No Issue Provided!") As ?issue)}}
                        
                        OPTIONAL {{?idraw wdt:{page(s)} ?pageraw.
                                   BIND(COALESCE(?pageraw, "No Pages Provided!") As ?page)}}
                        }}
   
                 GROUP BY ?id ?label ?description ?entrytypelabel ?journalInfo ?languagelabel ?title ?date ?volume ?issue ?page ''' },

  'Wikidata': {

       'QID_FULL': '''SELECT ?label ?description ?doi         
                        (GROUP_CONCAT(DISTINCT(?authorInfo); separator=" | ") AS ?authorInfos)                                                        
                        ?entrytypelabel ?journalInfo ?languagelabel                               
                        ?title ?date ?volume ?issue ?page          

                 WHERE {{ VALUES ?publication {{ wd:{0} }}
                 
                        OPTIONAL {{ ?publication wdt:P356 ?doi. }}

                        OPTIONAL {{?publication rdfs:label ?labelraw.
                                  FILTER (lang(?labelraw) = 'en')}}
                        BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)
                        
                        OPTIONAL {{?publication schema:description ?descriptionraw.
                                  FILTER (lang(?descriptionraw) = 'en')}}
                        BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)
                        
                        OPTIONAL {{?publication (wdt:P50 | wdt:P2093) ?authorraw.
                                  BIND(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorraw), STR(wd:))), "") AS ?author)
     
                                  OPTIONAL {{?authorraw rdfs:label ?authorlabelraw.
                                            FILTER (lang(?authorlabelraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorlabelraw, ?authorraw), "") AS ?authorlabel)
   
                                  OPTIONAL {{?authorraw schema:description ?authordescriptionraw.
                                            FILTER (lang(?authordescriptionraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authordescriptionraw, ""), "") AS ?authordescription)
   
                                  OPTIONAL {{?authorraw wdt:P496 ?authororcidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authororcidraw, ""), "") AS ?authororcid)
                          
                                  OPTIONAL {{?authorraw wdt:P ?authorwikidataidraw}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorwikidataidraw), STR(wd:))), ""), "") AS ?authorwikidataid)
   
                                  OPTIONAL {{?authorraw wdt:P1556 ?authorzbmathidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorzbmathidraw, ""), "") AS ?authorzbmathid)
     
                                  BIND(CONCAT(?author, " <|> ", ?authorlabel, " <|> ", ?authordescription, " <|> ", ?authororcid, " <|> ", ?authorzbmathid, " <|> ", ?authorwikidataid) AS ?authorInfo)}}
   
                        OPTIONAL {{?publication wdt:P31 ?entrytyperaw.
                                  
                                   OPTIONAL {{?entrytyperaw rdfs:label ?entrytypelabelraw.
                                            FILTER (lang(?entrytypelabelraw) = 'en')}}
                                   BIND(COALESCE(?entrytypelabelraw, "No Label Provided!") As ?entrytypelabel)
                                  }}
                        
                        OPTIONAL {{?publication wdt:P1433 ?journalraw.
                                  BIND(CONCAT("wikidata:", STRAFTER(STR(?journalraw), STR(wd:))) AS ?journal)
                                  
                                  OPTIONAL {{?journalraw rdfs:label ?journallabelraw.
                                            FILTER (lang(?journallabelraw) = 'en')}}
                                  BIND(COALESCE(?journallabelraw, "No Label Provided!") As ?journallabel)
                                  
                                  OPTIONAL {{?journalraw schema:description ?journaldescriptionraw.
                                            FILTER (lang(?journaldescriptionraw) = 'en')}}
                                  BIND(COALESCE(?journaldescriptionraw, "No Description Provided!") As ?journaldescription)
                                  
                                  BIND(concat(?journal, " <|> ", ?journallabel, " <|> ", ?journaldescription) AS ?journalInfo)}}
                        
                        OPTIONAL {{?publication wdt:P407 ?languageraw.
                                   
                                   OPTIONAL {{?languageraw rdfs:label ?languagelabelraw.
                                              FILTER (lang(?languagelabelraw) = 'en')}}
                                   BIND(COALESCE(?languagelabelraw, "No Label Provided!") As ?languagelabel)}}
                        
                        OPTIONAL {{?publication wdt:P1476 ?titleraw.
                                   BIND(COALESCE(?titleraw, "No Title Provided!") As ?title)}}
                        
                        OPTIONAL {{?publication wdt:P577 ?dateraw.
                                   BIND(COALESCE(?dateraw, "No Publication Date Provided!") As ?date)}}
                        
                        OPTIONAL {{?publication wdt:P478 ?volumeraw.
                                   BIND(COALESCE(?volumeraw, "No Volume Provided!") As ?volume)}}
                        
                        OPTIONAL {{?publication wdt:P433 ?issueraw.
                                   BIND(COALESCE(?issueraw, "No Issue Provided!") As ?issue)}}
                        
                        OPTIONAL {{?publication wdt:P304 ?pageraw.
                                   BIND(COALESCE(?pageraw, "No Pages Provided!") As ?page)}}
                        }}
   
                 GROUP BY ?doi ?label ?description ?entrytypelabel ?journalInfo ?languagelabel ?title ?date ?volume ?issue ?page ''',

       'DOI_FULL': '''SELECT ?id ?label ?description         
                        (GROUP_CONCAT(DISTINCT(?authorInfo); separator=" | ") AS ?authorInfos)                                                        
                        ?entrytypelabel ?journalInfo ?languagelabel                               
                        ?title ?date ?volume ?issue ?page          

                 WHERE {{?idraw wdt:P356 "{0}".
                        BIND(CONCAT("wikidata:", STRAFTER(STR(?idraw), STR(wd:))) AS ?id)

                        OPTIONAL {{?idraw rdfs:label ?labelraw.
                                  FILTER (lang(?labelraw) = 'en')}}
                        BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)
                        
                        OPTIONAL {{?idraw schema:description ?descriptionraw.
                                  FILTER (lang(?descriptionraw) = 'en')}}
                        BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)
                        
                        OPTIONAL {{?idraw (wdt:P50 | wdt:P2093) ?authorraw.
                                  BIND(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorraw), STR(wd:))), "") AS ?author)
     
                                  OPTIONAL {{?authorraw rdfs:label ?authorlabelraw.
                                            FILTER (lang(?authorlabelraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorlabelraw, ?authorraw), "") AS ?authorlabel)
   
                                  OPTIONAL {{?authorraw schema:description ?authordescriptionraw.
                                            FILTER (lang(?authordescriptionraw) = 'en')}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authordescriptionraw, ""), "") AS ?authordescription)
   
                                  OPTIONAL {{?authorraw wdt:P496 ?authororcidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authororcidraw, ""), "") AS ?authororcid)
                          
                                  OPTIONAL {{?authorraw wdt:P ?authorwikidataidraw}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), CONCAT("wikidata:", STRAFTER(STR(?authorwikidataidraw), STR(wd:))), ""), "") AS ?authorwikidataid)
   
                                  OPTIONAL {{?authorraw wdt:P1556 ?authorzbmathidraw.}}
                                  BIND(COALESCE(IF(CONTAINS(STR(?authorraw), STR(wd:)), ?authorzbmathidraw, ""), "") AS ?authorzbmathid)
     
                                  BIND(CONCAT(?author, " <|> ", ?authorlabel, " <|> ", ?authordescription, " <|> ", ?authororcid, " <|> ", ?authorzbmathid, " <|> ", ?authorwikidataid) AS ?authorInfo)}}
   
                        OPTIONAL {{?idraw wdt:P31 ?entrytyperaw.
                                  
                                   OPTIONAL {{?entrytyperaw rdfs:label ?entrytypelabelraw.
                                            FILTER (lang(?entrytypelabelraw) = 'en')}}
                                   BIND(COALESCE(?entrytypelabelraw, "No Label Provided!") As ?entrytypelabel)
                                  }}
                        
                        OPTIONAL {{?idraw wdt:P1433 ?journalraw.
                                  BIND(CONCAT("wikidata:", STRAFTER(STR(?journalraw), STR(wd:))) AS ?journal)
                                  
                                  OPTIONAL {{?journalraw rdfs:label ?journallabelraw.
                                            FILTER (lang(?journallabelraw) = 'en')}}
                                  BIND(COALESCE(?journallabelraw, "No Label Provided!") As ?journallabel)
                                  
                                  OPTIONAL {{?journalraw schema:description ?journaldescriptionraw.
                                            FILTER (lang(?journaldescriptionraw) = 'en')}}
                                  BIND(COALESCE(?journaldescriptionraw, "No Description Provided!") As ?journaldescription)
                                  
                                  BIND(concat(?journal, " <|> ", ?journallabel, " <|> ", ?journaldescription) AS ?journalInfo)}}
                        
                        OPTIONAL {{?idraw wdt:P407 ?languageraw.
                                   
                                   OPTIONAL {{?languageraw rdfs:label ?languagelabelraw.
                                              FILTER (lang(?languagelabelraw) = 'en')}}
                                   BIND(COALESCE(?languagelabelraw, "No Label Provided!") As ?languagelabel)}}
                        
                        OPTIONAL {{?idraw wdt:P1476 ?titleraw.
                                   BIND(COALESCE(?titleraw, "No Title Provided!") As ?title)}}
                        
                        OPTIONAL {{?idraw wdt:P577 ?dateraw.
                                   BIND(COALESCE(?dateraw, "No Publication Date Provided!") As ?date)}}
                        
                        OPTIONAL {{?idraw wdt:P478 ?volumeraw.
                                   BIND(COALESCE(?volumeraw, "No Volume Provided!") As ?volume)}}
                        
                        OPTIONAL {{?idraw wdt:P433 ?issueraw.
                                   BIND(COALESCE(?issueraw, "No Issue Provided!") As ?issue)}}
                        
                        OPTIONAL {{?idraw wdt:P304 ?pageraw.
                                   BIND(COALESCE(?pageraw, "No Pages Provided!") As ?page)}}
                        }}
   
                 GROUP BY ?id ?label ?description ?entrytypelabel ?journalInfo ?languagelabel ?title ?date ?volume ?issue ?page ''' },

       'authors': '''SELECT (GROUP_CONCAT(DISTINCT(?authorInfo); separator=" | ") AS ?authorInfos)
                          WHERE {{
                            {{
                              ?authorraw wdt:{2} ?orcid .
                              BIND(STRAFTER(STR(?authorraw), STR(wd:)) AS ?author)
                              VALUES ?orcid {{{0}}}
                            }}
                            UNION
                            {{
                              ?authorraw wdt:{3} ?zbmathid .
                              BIND(STRAFTER(STR(?authorraw), STR(wd:)) AS ?author)
                              VALUES ?zbmathid {{{1}}}
                            }}

                            OPTIONAL {{?authorraw rdfs:label ?labelraw.
                                       FILTER (lang(?labelraw) = 'en')}}
                            BIND(COALESCE(?labelraw, "No Label Provided!") AS ?label)

                            OPTIONAL {{?authorraw schema:description ?descriptionraw.
                                       FILTER (lang(?descriptionraw) = 'en')}}
                            BIND(COALESCE(?descriptionraw, "No Description Provided!") AS ?description)

                            OPTIONAL {{?authorraw wdt:{4} ?wikidataidraw}}
                            BIND(COALESCE(?wikidataidraw, "") AS ?wikidataid)
                            
                            OPTIONAL {{?authorraw wdt:{2} ?ORCIDRAW}}
                            BIND(COALESCE(?ORCIDRAW, "") AS ?ORCID)
                            
                            OPTIONAL {{?authorraw wdt:{3} ?ZBMATHIDRAW}}
                            BIND(COALESCE(?ZBMATHIDRAW, "") AS ?ZBMATHID)

                            BIND(CONCAT(?author, " <|> ", ?label, " <|> ", ?description, " <|> ", ?ORCID, " <|> ", ?ZBMATHID, " <|> ", ?wikidataid) AS ?authorInfo)

                            }}''',

       'journal': '''SELECT ?journalInfos
                           WHERE {{
                               
                               ?journalraw wdt:{1} "{0}".
                               BIND(STRAFTER(STR(?journalraw),STR(wd:)) AS ?journal).
                               
                               OPTIONAL {{?journalraw rdfs:label ?journallabelraw.
                                         FILTER (lang(?journallabelraw) = 'en')}}
                               BIND(COALESCE(?journallabelraw, "No Label Provided!") As ?journallabel)
                               
                               OPTIONAL {{?journalraw schema:description ?journaldescriptionraw.
                                         FILTER (lang(?journaldescriptionraw) = 'en')}}
                               BIND(COALESCE(?journaldescriptionraw, "No Description Provided!") As ?journaldescription)
                               
                               BIND(concat(?journal, " <|> ", ?journallabel, " <|> ", ?journaldescription) AS ?journalInfos)
 
                             }}''',

      'PublicationMathAlgoDB': '''PREFIX pb: <https://mardi4nfdi.de/mathalgodb/0.1/publication#>
                                   PREFIX : <https://mardi4nfdi.de/mathalgodb/0.1#>
                                   PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                   PREFIX dc: <http://purl.org/spar/datacite/>

                                  SELECT DISTINCT ?id ?label ?description ?doi
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?appliesentity, " | ", ?appliesentitylabel, " | ", ?appliesentitydescription)); SEPARATOR=" / ") AS ?applies)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?analyzesentity, " | ", ?analyzesentitylabel, " | ", ?analyzesentitydescription)); SEPARATOR=" / ") AS ?analyzes)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?documentsentity, " | ", ?documentsentitylabel, " | ", ?documentsentitydescription, " | ", ?documentsentityclass)); SEPARATOR=" / ") AS ?documents)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?inventsentity, " | ", ?inventsentitylabel, " | ", ?inventsentitydescription)); SEPARATOR=" / ") AS ?invents)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?studiesentity, " | ", ?studiesentitylabel, " | ", ?studiesentitydescription)); SEPARATOR=" / ") AS ?studies)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?surveysentity, " | ", ?surveysentitylabel, " | ", ?surveysentitydescription)); SEPARATOR=" / ") AS ?surveys)
                                                  (GROUP_CONCAT(DISTINCT(CONCAT(?usesentity, " | ", ?usesentitylabel, " | ", ?usesentitydescription, " | ", ?usesentityclass)); SEPARATOR=" / ") AS ?uses)
                                                                                    
                                  WHERE {{
                                          VALUES ?idraw {{ pb:{0} }}

                                          ?idraw a :publication.
                                          BIND(CONCAT("mathalgodb:", STRAFTER(STR(?idraw), "#")) AS ?id)
    
                                          OPTIONAL {{?idraw dc:hasIdentifier ?doiraw 
                                                        BIND(STRAFTER(STR(?doiraw), ":") AS ?doi) }}

                                          OPTIONAL {{?idraw rdfs:label ?labelraw.}}
                                          BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)

                                          OPTIONAL {{?idraw rdfs:comment ?descriptionraw.}}
                                          BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)

                                          OPTIONAL {{?idraw :applies ?appliesentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?appliesentityraw), "#")) AS ?appliesentity)

                                                     OPTIONAL {{?appliesentityraw rdfs:label ?appliesentitylabelraw.}}
                                                     BIND(COALESCE(?appliesentitylabelraw, "No Label Provided!") As ?appliesentitylabel)

                                                     OPTIONAL {{?appliesentityraw rdfs:comment ?appliesentitydescriptionraw.}}
                                                     BIND(COALESCE(?appliesentitydescriptionraw, "No Description Provided!") As ?appliesentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :analyzes ?analyzesentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?analyzesentityraw), "#")) AS ?analyzesentity)

                                                     OPTIONAL {{?analyzesentityraw rdfs:label ?analyzesentitylabelraw.}}
                                                     BIND(COALESCE(?analyzesentitylabelraw, "No Label Provided!") As ?analyzesentitylabel)

                                                     OPTIONAL {{?analyzesentityraw rdfs:comment ?analyzesentitydescriptionraw.}}
                                                     BIND(COALESCE(?analyzesentitydescriptionraw, "No Description Provided!") As ?analyzesentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :documents ?documentsentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?documentsentityraw), "#")) AS ?documentsentity)

                                                     ?documentsentityraw a ?documentsentityclassraw.
                                                     BIND(STRAFTER(STR(?documentsentityclassraw), "#") AS ?documentsentityclass)
                                                     FILTER (?documentsentityclassraw IN (:benchmark, :software))

                                                     OPTIONAL {{?documentsentityraw rdfs:label ?documentsentitylabelraw.}}
                                                     BIND(COALESCE(?documentsentitylabelraw, "No Label Provided!") As ?documentsentitylabel)

                                                     OPTIONAL {{?documentsentityraw rdfs:comment ?documentsentitydescriptionraw.}}
                                                     BIND(COALESCE(?documentsentitydescriptionraw, "No Description Provided!") As ?documentsentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :invents ?inventsentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?inventsentityraw), "#")) AS ?inventsentity)

                                                     OPTIONAL {{?inventsentityraw rdfs:label ?inventsentitylabelraw.}}
                                                     BIND(COALESCE(?inventsentitylabelraw, "No Label Provided!") As ?inventsentitylabel)

                                                     OPTIONAL {{?inventsentityraw rdfs:comment ?inventsentitydescriptionraw.}}
                                                     BIND(COALESCE(?inventsentitydescriptionraw, "No Description Provided!") As ?inventsentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :studies ?studiesentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?studiesentityraw), "#")) AS ?studiesentity)

                                                     OPTIONAL {{?studiesentityraw rdfs:label ?studiesentitylabelraw.}}
                                                     BIND(COALESCE(?studiesentitylabelraw, "No Label Provided!") As ?studiesentitylabel)

                                                     OPTIONAL {{?studiesentityraw rdfs:comment ?studiesentitydescriptionraw.}}
                                                     BIND(COALESCE(?studiesentitydescriptionraw, "No Description Provided!") As ?studiesentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :reviews ?surveysentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?surveysentityraw), "#")) AS ?surveysentity)

                                                     OPTIONAL {{?surveysentityraw rdfs:label ?surveysentitylabelraw.}}
                                                     BIND(COALESCE(?surveysentitylabelraw, "No Label Provided!") As ?surveysentitylabel)

                                                     OPTIONAL {{?surveysentityraw rdfs:comment ?surveysentitydescriptionraw.}}
                                                     BIND(COALESCE(?surveysentitydescriptionraw, "No Description Provided!") As ?surveysentitydescription)
                                                   }}

                                          OPTIONAL {{?idraw :uses ?usesentityraw.
                                                     BIND(CONCAT("mathalgodb:", STRAFTER(STR(?usesentityraw), "#")) AS ?usesentity)

                                                     ?usesentityraw a ?usesentityclassraw.
                                                     BIND(STRAFTER(STR(?usesentityclassraw), "#") AS ?usesentityclass)
                                                     FILTER (?usesentityclass IN (:benchmark, :software))

                                                     OPTIONAL {{?usesentityraw rdfs:label ?usesentitylabelraw.}}
                                                     BIND(COALESCE(?usesentitylabelraw, "No Label Provided!") As ?usesentitylabel)

                                                     OPTIONAL {{?usesentityraw rdfs:comment ?usesentitydescriptionraw.}}
                                                     BIND(COALESCE(?usesentitydescriptionraw, "No Description Provided!") As ?usesentitydescription)
                                                   }}
                                        }}
                                  GROUP BY ?id ?label ?description ?doi''',

                'PublicationMathAlgoDBDOI': '''PREFIX pb: <https://mardi4nfdi.de/mathalgodb/0.1/publication#>
                                   PREFIX : <https://mardi4nfdi.de/mathalgodb/0.1#>
                                   PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                   PREFIX dc: <http://purl.org/spar/datacite/>

                                  SELECT DISTINCT ?id ?label ?description
                                                                                    
                                  WHERE {{
                                          ?idraw a :publication.
                                          ?idraw dc:hasIdentifier ?doi
                                          FILTER(LCASE(STR(?doi)) = LCASE("doi:{0}"))

                                          BIND(CONCAT("mathalgodb:", STRAFTER(STR(?idraw), "#")) AS ?id)
    
                                          OPTIONAL {{?idraw rdfs:label ?labelraw.}}
                                          BIND(COALESCE(?labelraw, "No Label Provided!") As ?label)

                                          OPTIONAL {{?idraw rdfs:comment ?descriptionraw.}}
                                          BIND(COALESCE(?descriptionraw, "No Description Provided!") As ?description)


                                        }}
                                  GROUP BY ?id ?label ?description'''
                                          }


import html

from ..config import endpoint
from ..getters import get_general_item_url, get_items, get_properties
from ..queries import query_sparql
from .sparql import algorithmic_problem_sparql, software_sparql, algorithmic_problem_filter_sparql, quote_sparql, quantity_sparql, quantity_filter_sparql, task_sparql, formulation_sparql, res_obj_sparql, res_disc_sparql, mmsio_sparql, query_base, query_base_model, query_base_algorithm, problem_sparql, problem_filter_sparql, field_sparql

def search(answers, options):
    
    if answers['search'].get('options') == options['InterdisciplinaryWorkflow']:

        # SPARQL via Research Objectives

        quote_str = ''
        res_obj_strs = ''
        
        # If SPARQL query via research objective desired
        if answers['search'].get('via_research_objective') == options['Yes']:
            quote_str = quote_sparql
            # Separate key words for SPARQL query vie research objective
            if answers['search'].get('research_objective'):
                for res_obj in answers['search']['research_objective'].values():
                    # Define Filters for SPARQL queries
                    res_obj_strs += res_obj_sparql.format(res_obj.lower())

        # SPARQL via Research Disciplines

        res_disc_str = ''

        # If SPARQL query via research discipline desired
        if answers['search'].get('via_research_discipline') == options['Yes']:
            # Separate disciplines for SPARQL query via research discipline 
            if answers['search'].get('research_discipline'):
                for key in answers['search']['research_discipline'].keys():
                    # Get ID and Name of Research Discipline
                    ID = answers['search']['research_discipline'][key]['ID'].split(':')[1]
                    Name = answers['search']['research_discipline'][key]['Name']
                    answers['search']['research_discipline'][key].update({'ID': ID})
                    answers['search']['research_discipline'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    res_disc_str += res_disc_sparql.format(ID.split(':')[-1], **get_items(), **get_properties())

        # SPARQL via Mathematical Models, Methods, Softwares, Input or Output Data Sets

        mmsios_str = ''

        # If SPARQL query via Mathematical Models, Methods, Softwares, Input or Output Data Sets
        if answers['search'].get('via_workflow_entity') == options['Yes']:
            # Separate Mathematical Model, Methods, Software, Input or Output Data Sets
            if answers['search'].get('workflow_entity'):
                for key in answers['search']['workflow_entity'].keys():
                    # Get ID and Name of Research Discipline
                    ID = answers['search']['workflow_entity'][key]['ID'].split(':')[1]
                    Name = answers['search']['workflow_entity'][key]['Name']
                    answers['search']['workflow_entity'][key].update({'ID': ID})
                    answers['search']['workflow_entity'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    mmsios_str += mmsio_sparql.format(ID.split(':')[-1], **get_items(), **get_properties())

        # Set up entire SPARQL query
        query = "\n".join(line for line in query_base.format(res_disc_str, mmsios_str, quote_str, res_obj_strs, **get_items(), **get_properties()).splitlines() if line.strip())
        
        # Add Query to answer dictionary
        answers['query'] = html.escape(query).replace('\n', '<br>')

        # Query MaRDI Portal
        results = query_sparql(query, endpoint['mardi']['sparql'])

        # Number of Results
        answers['no_results'] = str(len(results))

        # Generate Links to Wikipage and Knowledge Graoh Entry of Results
        links=[]
        for result in results:
            links.append([result["label"]["value"], f"{endpoint['mardi']['uri']}/wiki/workflow:{result['qid']['value'][1:]}", f"{get_general_item_url()}{result['qid']['value']}"])

        answers['links'] = links

    elif answers['search'].get('options') == options['MathematicalModel']:

        # SPARQL via Research Problems

        pro_str = ''
        pro_fil_strs = ''
        
        # If SPARQL query via research objective desired
        if answers['search'].get('via_research_problem') == options['Yes']:
            pro_str = problem_sparql.format(**get_items(), **get_properties())
            # Separate key words for SPARQL query vie research objective
            if answers['search'].get('research_problem'):
                for res_pro in answers['search']['research_problem'].values():
                    # Define Filters for SPARQL queries
                    pro_fil_strs += problem_filter_sparql.format(res_pro.lower())

        # SPARQL via Research Fields
        
        fie_str = ''

        # If SPARQL query via research field desired
        if answers['search'].get('via_research_field') == options['Yes']:
            #fie_str = field_sparql
            # Separate key words for SPARQL query vie research objective
            if answers['search'].get('research_field'):
                for key in answers['search']['research_field'].keys():
                    # Get ID and Name of Research Field
                    ID = answers['search']['research_field'][key]['ID'].split(':')[1]
                    Name = answers['search']['research_field'][key]['Name']
                    answers['search']['research_field'][key].update({'ID': ID})
                    answers['search']['research_field'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    fie_str += field_sparql.format(ID, **get_items(), **get_properties())

        # SPARQL via Mathematical Formulations, Computational Task and Quantities

        for_str = ''
        ta_str = ''
        qu_str = quantity_sparql.format(**get_items(), **get_properties())
        
        # If SPARQL query via model entity desired
        if answers['search'].get('via_model_entity') == options['Yes']:
            # Via Formulations
            if answers['search'].get('model_formulation'):
                for key in answers['search']['model_formulation'].keys():
                    # Get ID and Name of Formulation
                    ID = answers['search']['model_formulation'][key]['ID'].split(':')[1]
                    Name = answers['search']['model_formulation'][key]['Name']
                    answers['search']['model_formulation'][key].update({'ID': ID})
                    answers['search']['model_formulation'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    for_str += formulation_sparql.format(ID, **get_items(), **get_properties())
            # Via Computational Tasls
            if answers['search'].get('model_task'):
                for key in answers['search']['model_task'].keys():
                    # Get ID and Name of Computational Task
                    ID = answers['search']['model_task'][key]['ID'].split(':')[1]
                    Name = answers['search']['model_task'][key]['Name']
                    answers['search']['model_task'][key].update({'ID': ID})
                    answers['search']['model_task'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    ta_str += task_sparql.format(ID, **get_items(), **get_properties())
            # Via Computational Tasls
            if answers['search'].get('model_quantity'):
                for idx, key in enumerate(answers['search']['model_quantity'].keys()):
                    # Get ID and Name of Computational Task
                    ID = answers['search']['model_quantity'][key]['ID'].split(':')[1]
                    Name = answers['search']['model_quantity'][key]['Name']
                    answers['search']['model_quantity'][key].update({'ID': ID})
                    answers['search']['model_quantity'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    if idx == 0:
                        qu_str += quantity_filter_sparql.format(ID, **get_items(), **get_properties())
                    else:
                        qu_str += """\n UNION""" + quantity_filter_sparql.format(ID, **get_items(), **get_properties()) 

        # Set up entire SPARQL query
        query = "\n".join(line for line in query_base_model.format(pro_str, pro_fil_strs, fie_str, for_str, ta_str, qu_str, **get_items(), **get_properties()).splitlines() if line.strip())

        # Add Query to answer dictionary
        answers['query'] = html.escape(query).replace('\n', '<br>')

        # Query MathModDB
        results = query_sparql(query, endpoint['mardi']['sparql'])

        # Number of Results
        answers['no_results'] = str(len(results))

        # Generate Links to Entry
        links=[]
        for result in results:
            links.append([result["label"]["value"], f"{endpoint['mardi']['uri']}/wiki/model:{result['qid']['value'][1:]}", f"{get_general_item_url()}{result['qid']['value']}"])
            
        answers['links'] = links

    elif answers['search'].get('options') == options['Algorithm']:

        # SPARQL via Algorithmic Problems

        apr_str = ''
        apr_fil_strs = ''
        
        # If SPARQL query via research objective desired
        if answers['search'].get('via_algorithmic_problem') == options['Yes']:
            apr_str = algorithmic_problem_sparql
            # Separate key words for SPARQL query vie research objective
            if answers['search'].get('algorithmic_problem'):
                for alg_pro in answers['search']['algorithmic_problem'].values():
                    # Define Filters for SPARQL queries
                    apr_fil_strs += algorithmic_problem_filter_sparql.format(alg_pro.lower())

        # SPARQL via Softwares
        
        sof_str = ''
        
        # If SPARQL query via software desired
        if answers['search'].get('via_software') == options['Yes']:
            # Separate key words for SPARQL query vie research objective
            if answers['search'].get('software'):
                for key in answers['search']['software'].keys():
                    # Get ID and Name of Software
                    ID = answers['search']['software'][key]['ID'].split(':')[1]
                    Name = answers['search']['software'][key]['Name']
                    answers['search']['software'][key].update({'ID': ID})
                    answers['search']['software'][key].update({'Name': Name})
                    # Define Filters for SPARQL queries
                    sof_str += software_sparql.format(f"software:{ID}")

        # Set up entire SPARQL query
        query = "\n".join(line for line in query_base_algorithm.format(apr_str, apr_fil_strs, sof_str).splitlines() if line.strip())
        
        # Add Query to answer dictionary
        answers['query'] = html.escape(query).replace('\n', '<br>')

        # Query MathAlgoDB
        results = query_sparql(query, endpoint['mathalgodb']['sparql'])

        # Number of Results
        answers['no_results'] = str(len(results))

        # Generate Links to Entry
        links=[]
        for result in results:
            links.append([result["label"]["value"], endpoint['mathalgodb']['uri'] + 'object/al:' + result["qid"]["value"], ''])

        answers['links'] = links

    return answers

    



'''General Constants used in MaRDMO'''

from . import rules

flag_dict = {
             (False, False, False, False, False): rules.rule_0,
             (True, False, False, False, False): rules.rule_1,
             (False, True, False, False, False): rules.rule_2,
             (True, True, False, False, False): rules.rule_3,
             (False, True, True, False, False): rules.rule_4,
             (True, False, True, False, False): rules.rule_5,
             (True, True, True, False, False): rules.rule_6,
             (True, False, False, False, True): rules.rule_7,
             (False, True, False, False, True): rules.rule_8,
             (False, False, False, True, False): rules.rule_9,
             (False, False, True, False, False): rules.rule_10,
             (False, False, True, True, False): rules.rule_11,
             (False, True, False, True, False): rules.rule_12,
             (True, False, False, True, False): rules.rule_13,
             (True, True, False, True, False): rules.rule_14,
             (True, False, True, True, False): rules.rule_15,
             (True, True, True, True, False): rules.rule_16,
}



'''Module routing all post_save and post_delete handlers.'''

from django.db.models.signals import post_save
from django.dispatch import receiver
from rdmo.projects.models import Value
from .builders import build_handler_map

HANDLER_MAP = build_handler_map()

@receiver(post_save, sender=Value)
def mardmo_router(sender, instance, update_fields=None, **kwargs):
    """Global post_save router for MaRDMO plugin: dispatch Value saves 
       to correct handler."""

    if not instance:
        return

    catalog = getattr(instance.project, "catalog", None)
    if not catalog or not str(catalog).endswith(
        ("mardmo-model-catalog",
         "mardmo-algorithm-catalog",
         "mardmo-interdisciplinary-workflow-catalog")
    ):
        return

    attr_uri = getattr(instance.attribute, "uri", None)
    if not attr_uri:
        return

    catalog_name = str(catalog).rsplit("/", maxsplit = 1)[-1]

    handler = HANDLER_MAP.get(catalog_name, {}).get(attr_uri)
    if handler:
        handler(instance)



from dataclasses import asdict

from .sparql import queryPreview
from .models import ModelProperties, Variables, Parameters
from .constants import REPRODUCIBILITY

from ..config import endpoint
from ..getters import get_items, get_options, get_properties
from ..helpers import unique_items
from ..queries import query_sparql
from ..payload import GeneratePayload

class prepareWorkflow:

    def __init__(self):
        self.items = get_items()
        self.properties = get_properties()

    def preview(self, data):

        # Update Model Properties via MathModDB
        if data.get('model',{}).get('ID'):
            _, id = data['model']['ID'].split(':')
            query = queryPreview['basic'].format(id, **self.items, **self.properties)
            basic = query_sparql(query, endpoint['mardi']['sparql'])
            if basic:
                data.get('model', {}).update(asdict(ModelProperties.from_query(basic)))
        
        # Update Model Variables and Parameters via MathModDB
        if data.get('model', {}).get('task'):            
            query = queryPreview['variables'].format(' '.join(f"wd:{value.get('ID', '').split(':')[1]}" for _, value in data['model']['task'].items()), **self.items, **self.properties)
            variables = query_sparql(query, endpoint['mardi']['sparql'])
            if variables:
                for idx, variable in enumerate(variables):
                    data.setdefault('variables', {}).update({idx: asdict(Variables.from_query(variable))})
            query = queryPreview['parameters'].format(' '.join(f"wd:{value.get('ID', '').split(':')[1]}" for _, value in data['model']['task'].items()), **self.items, **self.properties)
            parameters = query_sparql(query, endpoint['mardi']['sparql'])
            if parameters:
                for idx, parameter in enumerate(parameters):
                    data.setdefault('parameters', {}).update({idx: asdict(Parameters.from_query(parameter))})

        return data

    def export(self,data, title, url):

        items = unique_items(data, title)
        
        payload = GeneratePayload(url, items)

        # Load Options
        options = get_options()

        # Add / Retrieve Components of Interdisciplinary Workflow Item
        payload.process_items()
        
        ### Add additional Algorithms / Methods Information
        for method in data.get('method', {}).values():

            # Continue if no ID exists
            if not method.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(method)

            # Add Class
            if 'mathalgodb' in method['ID']:
                payload.add_answer(
                    verb=self.properties['instance of'],
                    object_and_type=[
                        self.items['algorithm'],
                        'wikibase-item',
                    ]
                )
            else:
                payload.add_answer(
                    verb=self.properties['instance of'],
                    object_and_type=[
                        self.items['method'],
                        'wikibase-item',
                    ]
                )
                
        ### Add additional Software Information
        for software in data.get('software', {}).values():

            # Continue if no ID exists
            if not software.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(software)

            # Add Class
            payload.add_answer(
                    verb=self.properties['instance of'],
                    object_and_type=[
                        self.items['software'],
                        'wikibase-item',
                    ]
                )
            
            # Add References of the Software
            for reference in software.get('Reference', {}).values():
                if reference[0] == options['DOI']:
                    payload.add_answer(
                        verb=self.properties['DOI'],
                        object_and_type=[
                            reference[1],
                            'external-id',
                        ]
                    )
                elif reference[0] == options['SWMATH']:
                    payload.add_answer(
                        verb=self.properties['swMath work ID'],
                        object_and_type=[
                            reference[1],
                            'external-id',
                        ]
                    )
                elif reference[0] == options['URL']:
                    payload.add_answer(
                        verb=self.properties['URL'],
                        object_and_type=[
                            reference[1],
                            'url',
                        ]
                    )
                    
            # Add Programming Languages
            payload.add_forward_relation_single(
                relation = self.properties['programmed in'],
                relatant = 'programminglanguage'
            )

            # Add Dependencies
            payload.add_forward_relation_single(
                relation = self.properties['depends on software'],
                relatant = 'dependency'
            )

            # Add Source Code Repository
            if software.get('Published', [''])[0] == options['YesText']:
                payload.add_answer(
                    verb=self.properties['source code repository URL'],
                    object_and_type=[
                        software['Published'][1],
                        'url',
                    ]
                )
                
            # Add Documentation / Manual
            if software.get('Documented', [''])[0] == options['YesText']:
                payload.add_answer(
                    verb=self.properties['user manual URL'],
                    object_and_type=[
                        software['Documented'][1],
                        'url',
                    ]
                )
                
        ### Add additional Hardware Information
        for hardware in data.get('hardware', {}).values():

            # Continue if no ID exists
            if not hardware.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(hardware)

            # Add Class
            payload.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    self.items['computer hardware'],
                    'wikibase-item',
                ]
            )
            
            # Add CPU
            payload.add_forward_relation_single(
                relation = self.properties['CPU'],
                relatant = 'cpu'
            )

            # Add Number of Computing Nodes
            if hardware['Nodes']:
                payload.add_answer(
                    verb=self.properties['has part(s)'], 
                    object_and_type=[
                        self.items['compute node'],
                        'wikibase-item',
                    ],
                    qualifier = [
                                    {
                                        "property": {
                                            "id": self.properties['quantity_property'],
                                        },
                                        "value": {
                                            "type": "value",
                                            "content": {
                                                "amount": f"+{hardware['Nodes']}",
                                                "unit": "1",
                                            },
                                        },
                                    }
                                ]
                )

            # Add Number of Processor Cores
            if hardware['Cores']:
                payload.add_answer(
                    verb=self.properties['number of processor cores'], 
                    object_and_type=[
                        {
                            "amount": f"+{hardware['Cores']}",
                            "unit":"1"
                        },
                        'quantity'
                    ]
                ) 

        ### Add additional Instrument Information
        for instrument in data.get('instrument', {}).values():

            # Continue if no ID exists
            if not instrument.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(instrument)

            # Add Class
            payload.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    self.items['research tool'],
                    'wikibase-item',
                ]
            )
            
            ### MORE INSTRUMENT INFORMATION TO ADD ###

        ### Add additional Dataset Information
        for dataset in data.get('dataset', {}).values():

            # Continue if no ID exists
            if not dataset.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(dataset)

            # Add Class
            payload.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    self.items['data set'],
                    'wikibase-item',
                ]
            )
            
            # Size of the data set
            if dataset.get('Size'):
                if dataset['Size'][0] == options['kilobyte']:
                    verb = self.properties['data size']
                    object = {"amount":f"+{dataset['Size'][1]}","unit": f"{endpoint['mardi']['uri']}/entity/{self.items['kilobyte']}"}
                elif dataset['Size'][0] == options['megabyte']:
                    verb = self.properties['data size']
                    object = {"amount":f"+{dataset['Size'][1]}","unit": f"{endpoint['mardi']['uri']}/entity/{self.items['megabyte']}"}
                elif dataset['Size'][0] == options['gigabyte']:
                    verb = self.properties['data size']
                    object = {"amount":f"+{dataset['Size'][1]}","unit": f"{endpoint['mardi']['uri']}/entity/{self.items['gigabyte']}"}
                elif dataset['Size'][0] == options['terabyte']:
                    verb = self.properties['data size']
                    object = {"amount":f"+{dataset['Size'][1]}","unit": f"{endpoint['mardi']['uri']}/entity/{self.items['terabyte']}"}
                elif dataset['Size'][0] == options['items']:
                    verb = self.properties['number of records']
                    object = {"amount":f"+{dataset['Size'][1]}","unit":"1"}

                payload.add_answer(
                    verb=verb,
                    object_and_type=[
                        object,
                        'quantity',
                    ]
                )
                
            # Add Data Type 
            payload.add_forward_relation_single(
                relation = self.properties['uses'], 
                relatant = 'datatype', 
                qualifier = payload.add_qualifier(
                    self.properties['object has role'],
                    'wikibase-item',
                    self.items['data type']
                )
            )

            # Add Representation Format 
            payload.add_forward_relation_single(
                relation = self.properties['uses'], 
                relatant = 'representationformat', 
                qualifier = payload.add_qualifier(
                    self.properties['object has role'],
                    'wikibase-item',
                    self.items['representation format']
                )
            )            

            # Add File Format
            if dataset.get('FileFormat'):
                payload.add_answer(
                    verb=self.properties['file extension'],
                    object_and_type=[
                        dataset['FileFormat'],
                        'string',
                    ]
                )
                
            # Add binary or text data
            if dataset.get('BinaryText'):
                if dataset['BinaryText'] == options['binary']:
                    payload.add_answer(
                        verb=self.properties['instance of'],
                        object_and_type=[
                            self.items['binary data'],
                            'wikibase-item',
                        ]
                    )
                elif dataset['BinaryText'] == options['text']:
                    payload.add_answer(
                        verb=self.properties['instance of'],
                        object_and_type=[
                            self.items['text data'],
                            'wikibase-item',
                        ]
                    )
                    
            # Data Set Proprietary
            if dataset.get('Proprietary'):
                if dataset['Proprietary'] == options['Yes']:
                    payload.add_answer(
                        verb=self.properties['instance of'],
                        object_and_type=[
                            self.items['proprietary information'],
                            'wikibase-item',
                        ]
                    )
                elif dataset['Proprietary'] == options['No']:
                    payload.add_answer(
                        verb=self.properties['instance of'],
                        object_and_type=[
                            self.items['open data'],
                            'wikibase-item',
                        ]
                    )
                    
            # Data Set to Publish
            if dataset.get('ToPublish'):
                if dataset['ToPublish'].get(0, ['',''])[0] == options['Yes']:
                    payload.add_answer(
                        verb=self.properties['mandates'],
                        object_and_type=[
                            self.items['data publishing'],
                            'wikibase-item',
                        ]
                    )
                    if dataset['ToPublish'].get(1, ['',''])[0] == options['DOI']:
                        payload.add_answer(
                            verb=self.properties['DOI'],
                            object_and_type=[
                                dataset['ToPublish'][1][1],
                                'external-id',
                            ]
                        )
                    if dataset['ToPublish'].get(2, ['',''])[0] == options['URL']:
                        payload.add_answer(
                            verb=self.properties['URL'],
                            object_and_type=[
                                dataset['ToPublish'][2][2],
                                'url',
                            ]
                        )
                        
            # Data Set To Archive
            if dataset.get('ToArchive'):
                if dataset['ToArchive'][0] == options['YesText']:
                    qualifier = []
                    if dataset['ToArchive'][1]:
                        qualifier = payload.add_qualifier(self.properties['end time'], 'time', {"time":f"+{dataset['ToArchive'][1]}-00-00T00:00:00Z","precision":9,"calendarmodel":"http://www.wikidata.org/entity/Q1985727"})
                    payload.add_answer(
                            verb=self.properties['mandates'],
                            object_and_type=[
                                self.items['research data archiving'],
                                'wikibase-item',
                            ],
                            qualifier=qualifier
                        )

        ### Add Process Step Information
        for processstep in data.get('processstep', {}).values():

            # Continue if no ID exists
            if not processstep.get('ID'):
                continue
            
            # Get Item Key
            payload.get_item_key(processstep)

            # Add Class
            payload.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    self.items['process step'],
                    'wikibase-item',
                ]
            )
            
            # Add Input Data Sets
            payload.add_forward_relation_single(
                relation = self.properties['input data set'],
                relatant = 'input'
            )

            # Add Output Data Sets
            payload.add_forward_relation_single(
                relation = self.properties['output data set'],
                relatant = 'output'
            )

            # Add applied Methods
            for method in processstep.get('method', {}).values():
                # Continue if no ID exists
                if not method.get('ID'):
                    continue
                # Get Entry Key
                method_item = payload.get_item_key(method, 'object')
                # Get Qualifier
                qualifier = []
                for parameter in method.get('Parameter', {}).values():
                    qualifier.extend(payload.add_qualifier(self.properties['comment'], 'string', parameter))
                # Add to Payload
                payload.add_answer(
                            verb=self.properties['uses'],
                            object_and_type=[
                                method_item,
                                'wikibase-item',
                            ],
                            qualifier=qualifier
                        )
                
            # Add Software Environment
            payload.add_forward_relation_single(
                relation = self.properties['platform'], 
                relatant = 'environmentSoftware',
                qualifier = payload.add_qualifier(
                    self.properties['object has role'],
                    'wikibase-item',
                    self.items['software']
                )
            )
            
            # Add Instrument Environment
            payload.add_forward_relation_single(
                relation = self.properties['platform'], 
                relatant = 'environmentInstrument',
                qualifier = payload.add_qualifier(
                    self.properties['object has role'],
                    'wikibase-item',
                    self.items['research tool']
                )
            )
            
            # Add Disciplines (math and non-math)
            for discipline in processstep.get('discipline', {}).values():
                # Check if new ID exists
                if 'msc:' in discipline.get('ID'):
                    _, id = discipline['ID'].split(':')
                    payload.add_answer(
                        verb=self.properties['MSC ID'],
                        object_and_type=[
                            id,
                            'external-id',
                        ]
                    )
                else:
                    # Get Discipline Key
                    discipline_item = payload.get_item_key(discipline, 'object')
                    # Add to Payload
                    payload.add_answer(
                        verb=self.properties['field of work'],
                        object_and_type=[
                            discipline_item,
                            'wikibase-item',
                        ]
                    )
                    
        for publication in data.get('publication', {}).values():

            # Continue if no ID exists
            if not publication.get('ID'):
                continue

            # Get Item Key
            payload.get_item_key(publication)

            if 'mardi' not in publication['ID'] and 'wikidata' not in publication['ID']:
                
                # Add the class of the Publication
                if publication.get('entrytype'):

                    if publication['entrytype'] == 'scholarly article':
                        pclass = self.items['scholarly article']
                    else:
                        pclass = self.items['publication']

                    payload.add_answer(
                        verb=self.properties['instance of'],
                        object_and_type=[
                            pclass,
                            'wikibase-item',
                        ]
                    )
               
                # Add the Title of the Publication
                if publication.get('title'):

                    payload.add_answer(
                        verb=self.properties['title'],
                        object_and_type=[
                            {"text": publication['title'], "language": "en"},
                            'monolingualtext',
                        ]
                    )

                # Add the Volume of the Publication
                if publication.get('volume'):

                    payload.add_answer(
                        verb=self.properties['volume'],
                        object_and_type=[
                            publication['volume'],
                            'string',
                        ]
                    )

                # Add the Issue of the Publication
                if publication.get('issue'):

                    payload.add_answer(
                        verb=self.properties['issue'],
                        object_and_type=[
                            publication['issue'],
                            'string',
                        ]
                    )

                # Add the Page(s) of the Publication
                if publication.get('page'):

                    payload.add_answer(
                        verb=self.properties['page(s)'],
                        object_and_type=[
                            publication['page'],
                            'string',
                        ]
                    )

                # Add the Date of the Publication
                if publication.get('date'):

                    payload.add_answer(
                        verb=self.properties['publication date'],
                        object_and_type=[
                            {"time":f"+{publication['date']}T00:00:00Z","precision":11,"calendarmodel":"http://www.wikidata.org/entity/Q1985727"},
                            'time',
                        ]
                    )

                # Add the DOI of the Publication
                if publication.get('reference', {}).get(0):

                    payload.add_answer(
                        verb=self.properties['DOI'],
                        object_and_type=[
                            publication['reference'][0][1],
                            'external-id',
                        ]
                    )
                
                # Add the Language of the Publication
                payload.add_forward_relation_single(
                    relation = self.properties['language of work or name'],
                    relatant = 'language'
                )

                # Add the Journal of the Publication
                payload.add_forward_relation_single(
                    relation = self.properties['published in'],
                    relatant = 'journal'
                )

                # Add the Authors of the Publication
                payload.add_forward_relation_single(
                    relation = self.properties['author'],
                    relatant = 'author',
                    alterantive = {
                        'relation': self.properties['author name string'], 
                        'relatant': 'Name'
                    }
                )
        
        # Add Interdisciplinary Workflow Information
        workflow = {'ID': 'not found', 'Name': title, 'Description': data.get('general', {}).get('objective')}
        
        # Get Item Key
        payload.get_item_key(workflow)

        # Add Class
        payload.add_answer(
            verb=self.properties['instance of'],
            object_and_type=[
                self.items['research workflow'],
                'wikibase-item',
            ]
        )
        
        # Procedure Description to Workflow
        if data.get('general', {}).get('procedure'):
            payload.add_answer(
                verb=self.properties['description'],
                object_and_type=[
                    data['general']['procedure'],
                    'string',
                ]
            )
            
        # Add Reproducibility Aspects
        for key, value in REPRODUCIBILITY.items():
            if data.get('reproducibility', {}).get(key) == options['Yes']:
                qualifier = []
                if data['reproducibility'].get(f'{key}condition'):
                    qualifier.extend(payload.add_qualifier(self.properties['comment'], 'string', data['reproducibility'][f'{key}condition']))
                payload.add_answer(
                    verb=self.properties['instance of'],
                    object_and_type=[
                        self.items[value],
                        'wikibase-item',
                    ],
                    qualifier=qualifier
                )
                
        # Add Transferability Aspects
        if data.get('reproducibility', {}).get('transferability'):
            qualifier = []
            for value in data['reproducibility']['transferability'].values():
                qualifier.extend(payload.add_qualifier(self.properties['comment'], 'string', value))
            payload.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    self.items['transferable research workflow'],
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add Model and Task the Workflow Uses
        if data.get('model'):
            #Continue if ID exists
            if data['model'].get('ID'):
                # Get Item Key
                model_item = payload.get_item_key(data['model'], 'object')
                # Add Statement with Qualifier
                qualifier = []
                for task in data['model'].get('task', {}).values():
                    qualifier.extend(payload.add_qualifier(self.properties['used by'], 'wikibase-item', payload.get_item_key(task, 'object')))
                payload.add_answer(
                    verb=self.properties['uses'],
                    object_and_type=[
                        model_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier
                )
            
        # Add Methods the Workflow Uses
        for value in data.get('method', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            method_item = payload.get_item_key(value, 'object')
            # Add Statement with Qualifier
            qualifier = []
            for parameter in value.get('Parameter', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['comment'], 'string', parameter))
            for software in value.get('software', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['implemented by'], payload.get_item_key(software, 'object')))
            for instrument in value.get('instrument', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['implemented by'], payload.get_item_key(instrument, 'object')))
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    method_item,
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add Software the Workflow uses
        for value in data.get('software', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            software_item = payload.get_item_key(value, 'object')
            # Add Statement with Qualifier
            qualifier = []
            for hardware in data.get('hardware', {}).values():
                for software in hardware.get('software', {}).values():
                    if (software.get('ID'), software.get('Name'), software.get('Description')) == (value['ID'], value['Name'], value['Description']):
                        hardware_item = payload.get_item_key(hardware, 'object')
                        qualifier.extend(payload.add_qualifier(self.properties['platform'], 'wikibase-item', hardware_item))
            if value.get('Version'):
                qualifier = payload.add_qualifier(self.properties['software version identifier'], 'string', value['Version'])
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    software_item,
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add Hardware the Workflow Uses
        for value in data.get('hardware', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            hardware_item = payload.get_item_key(value, 'object')
            # Add Satement with Qualifier
            qualifier = []
            for compiler in value.get('compiler', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['uses'], 'wikibase-item', payload.get_item_key(compiler, 'object')))
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    hardware_item,
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add instruments the workflow Uses
        for value in data.get('instrument', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            instrument_item = payload.get_item_key(value, 'object')
            # Add Statement with Qualifer
            qualifier = []
            if value.get('Version'):
                qualifier.extend(payload.add_qualifier(self.properties['edition number'], 'string', value['Version'])) 
            if value.get('SerialNumber'):
                qualifier.extend(payload.add_qualifier(self.properties['serial number'], 'string', value['SerialNumber'])) 
            for location in value.get('location', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['location'], 'wikibase-item', payload.get_item_key(location, 'object'))) 
            for software in value.get('software', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['uses'], 'wikibase-item', payload.get_item_key(software, 'object'))) 
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    instrument_item,
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add Data Sets the Workflow Uses
        for value in data.get('dataset', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            dataset_item = payload.get_item_key(value, 'object')
            # Add Statement
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    dataset_item,
                    'wikibase-item',
                ]
            )
            
        # Add Process Steps the Workflow Uses
        for value in data.get('processstep', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            processstep_item = payload.get_item_key(value, 'object')
            # Add Statement with Qualifier
            qualifier = []
            for parameter in value.get('parameter', {}).values():
                qualifier.extend(payload.add_qualifier(self.properties['comment'], 'string', parameter)) 
            payload.add_answer(
                verb=self.properties['uses'],
                object_and_type=[
                    processstep_item,
                    'wikibase-item',
                ],
                qualifier=qualifier
            )
            
        # Add Publications related to the Workflow
        for value in data.get('publication', {}).values():
            # Continue if no ID exists
            if not value.get('ID'):
                continue
            # Get Item Key
            publication_item = payload.get_item_key(value, 'object')
            # Add Statement
            payload.add_answer(
                verb=self.properties['cites work'],
                object_and_type=[
                    publication_item,
                    'wikibase-item',
                ]
            )
            
        # Construct Item Payloads
        payload.add_item_payload()

        # If Relations are added, check if they exist
        if any(key.startswith('RELATION') for key in payload.get_dictionary()):

            # Generate SPARQL Check Query
            query = payload.build_relation_check_query()
        
            # Perform Check Query for Relations
            check = query_sparql(query, endpoint['mardi']['sparql'])

            # Add Check Results
            payload.add_check_results(check)
            
        return payload.get_dictionary()



'''Functions to add Information to the Questionaire.'''

from rdmo.options.models import Option

from .config import BASE_URI
from .getters import get_id, get_options
from .helpers import (
    extract_parts,
    initialize_counter,
    process_qualifier,
    reduce_prefix,
    relation_exists,
    relevant_set_ids,
    value_editor,
)


def add_basics(project, text, questions, item_type, index = (None, None)):
    '''Function extracts Label, Description and Source of Items selected or defined
       in the ID Question on each Page. Label and Description are added to the Name
       and Description Questions on the individual Pages.
       
       Input: Selected/Defined Item as String 'Label (Description) [source]'
       Output: Label, Description and Source'''

    # Extract Label, Description, Source from ID Question
    label, description, source = extract_parts(text)

    # Add Label to Questionnaire
    value_editor(
        project = project,
        uri = f'{BASE_URI}{questions[item_type]["Name"]["uri"]}',
        info = {
            'text': label,
            'set_index': index[0],
            'set_prefix': index[1]
        }
    )

    # Add Description to Questionnaire
    value_editor(
        project = project,
        uri = f'{BASE_URI}{questions[item_type]["Description"]["uri"]}',
        info = {
            'text': description,
            'set_index': index[0],
            'set_prefix': index[1]
        }
    )

    return label, description, source

def add_entities(project, question_set, datas, source, prefix):
    '''Function checks if an Item selected in one Section of the Questionnaire is 
       defined in another Section of the Questionnaire. If the Item is not yet in
       the Questionnaire a new Page is created and the ID Question is answered.

       Input: Item Information
       Output: -'''

    # Generate ID, Name and Description URL from Set URL
    question = {'id': f'{question_set}/id',
                'name': f'{question_set}/name',
                'description': f'{question_set}/description'}

    # Get existing Set and Item Information
    info = {'set_ids': get_id(project, question_set, ['set_index']),
            'value_ids': get_id(project, question['id'], ['external_id']),
            'texts': get_id(project, question['id'], ['text']),
            'names': get_id(project, question['name'], ['text']),
            'descs': get_id(project, question['description'], ['text'])}

    # Add Item to Questionnaire
    idx = max(info['set_ids'], default = -1) + 1

    for data in datas:
        # Label Description String
        name_desc = f'{data.label} ({data.description})'
        # Check if Item already in Questionnaire via ID Question
        check_id = any(
            name_desc in text
            for text in info['texts']
            )
        # Check if Item already in Questionnaire via Name/Description Question
        check_name_desc = any(
            name_desc in f'{name} ({desc})'
            for name, desc in zip(info['names'], info['descs'])
            )
        # If Item not already in Questionnaire
        if data.id not in info['value_ids'] and not check_id and not check_name_desc:
            # Set up Page in Questionnaire
            value_editor(
                project = project,
                uri = question_set,
                info = {
                    'text': f"{prefix}{idx}",
                    'set_index': idx
                }
            )
            # Add ID Values
            value_editor(
                project = project,
                uri = question['id'],
                info = {
                    'text': f'{data.label} ({data.description}) [{source}]',
                    'external_id': f"{data.id}",
                    'set_index': idx
                }
            )

            # Update Index and existing Items
            idx += 1
            info['value_ids'].append(data.id)

def add_new_entities(project, question_set, datas, prefix):
    '''Function checks if an Item defined in one Section of the Questionnaire is 
       defined in another Section of the Questionnaire. If the Item is not yet in
       the Questionnaire a new Page is created and the ID Question is answered.

       Input: Item Information
       Output: -'''

    # Generate ID, Name and Description URL from Set URL
    question = {'id': f'{question_set}/id',
                'name': f'{question_set}/name',
                'description': f'{question_set}/description'}

    # Get existing Set and Item Information
    info = {'set_ids': get_id(project, question_set, ['set_index']),
            'names': get_id(project, question['name'], ['text']),
            'descs': get_id(project, question['description'], ['text'])}

    # Add Publication to Questionnaire
    idx = max(info['set_ids'], default = -1) + 1
    for data in datas:
        # Label Description String
        name_desc = f'{data.label} ({data.description})'
        # Check if Item already in Questionnaire via Name/Description Question
        check_name_desc = any(
            name_desc == f'{name} ({desc})'
            for name, desc in zip(info['names'], info['descs'])
            )
        # If Item not already in Questionnaire
        if not check_name_desc:
            # Set up Page
            value_editor(
                project = project,
                uri = question_set,
                info = {
                    'text': f"{prefix}{idx}",
                    'set_index': idx
                }
            )
            # Add ID Values
            value_editor(
                project = project,
                uri = question['id'],
                info = {
                    'text': 'not found',
                    'external_id': 'not found',
                    'set_index': idx
                }
            )
            # Add Name Values
            value_editor(
                project = project,
                uri = question['name'],
                info = {
                    'text': data.label,
                    'set_prefix': idx
                }
            )
            # Add Description Values
            value_editor(
                project = project,
                uri = question['description'],
                info = {
                    'text': data.description,
                    'set_prefix': idx
                }
            )

            # Update Index
            idx += 1

def add_relations_static(project, data, props, index, statement):
    '''Function checks if a related pair (relation and relatant) are part of the 
       Questionnaire (fixed relation). If the pair is not yet in the Questionnaire 
       it is added to the Questionnaire. Qualifiers are likewise checked and added.

       Input: Item Information + Relation (if flexible)
       Output: -'''

    # Get existing Set and Item Information
    info = {'set_prefix_ids': get_id(project, statement['relatant'], ['set_prefix']),
            'set_index_ids': get_id(project, statement['relatant'], ['set_index']),
            'collection_ids': get_id(project, statement['relatant'], ['collection_index']),
            'value_ids': get_id(project, statement['relatant'], ['external_id']),
            'texts': get_id(project, statement['relatant'], ['text'])}

    # Get reduced set_prefixes
    index.update({'set_prefix_reduced': reduce_prefix(index['set_prefix'])})

    # Set initial value of counter
    index.update({'idx': initialize_counter(info['collection_ids'])})

    # Add Relations and Relatants
    for prop in props['keys']:
        for value in getattr(data, prop):

            # Get Source and Label Description String
            source, _ = value.id.split(':')

            # Check if Relatant exists
            matches = relation_exists(
                value = value,
                set_prefix_red = index['set_prefix_reduced'],
                info = info)

            if matches:
                # Continue if existing
                continue

            # Add Relatant to Questionnaire
            value_editor(
                project = project,
                uri = statement['relatant'],
                info = {
                    'text': f"{value.label} ({value.description}) [{source}]",
                    'external_id': value.id,
                    'collection_index': index['idx'],
                    'set_index': 0,
                    'set_prefix': index['set_prefix']
                }
            )

            #Update Index
            index['idx'] += 1

            # Update existing IDs, Texts, and Relations
            info['value_ids'].append(value.id)
            info['set_prefix_ids'].append(index['set_prefix_reduced'])
            info['texts'].append(f"{value.label} ({value.description}) [{source}]")

def add_relations_flexible(project, data, props, index, statement):
    '''Function checks if an Item and a Relation (fixed/flexible) are part of the 
       Questionnaire. If the Item / Relation Pair is not yet in the Questionnaire 
       it is added to the Questionnaire. Qualifiers are likewise checked and added.

       Input: Item Information + Relation (if flexible)
       Output: -'''

    # Get existing Set, Item and Relation Information
    info = {'set_prefix_ids': get_id(project, statement['relatant'], ['set_prefix']),
            'set_index_ids': get_id(project, statement['relatant'], ['set_index']),
            'collection_ids': get_id(project, statement['relatant'], ['collection_index']),
            'value_ids': get_id(project, statement['relatant'], ['external_id']),
            'texts': get_id(project, statement['relatant'], ['text']),
            'rels': get_id(project, statement['relation'], ['option_uri'])}

    # Get reduced set prefix ids
    index.update({'set_prefix_reduced': reduce_prefix(index['set_prefix'])})

    # Get relevant set index ids
    ids = relevant_set_ids(info, index['set_prefix_reduced'])

    # Set initial value of counter
    index.update({'idx': initialize_counter(ids)})

    # Add Relations and Relatants
    for prop in props['keys']:
        for value in getattr(data, prop):

            # Get Source and Label Description String
            source, _ = value.id.split(':')

            # Check if Relation / Relatant Combination exists (flexible relation)
            matches = relation_exists(
                value = value,
                set_prefix_red = index['set_prefix_reduced'],
                info = info,
                relation_id = props['mapping'][prop])

            if matches:
                # Continue if existing
                continue

            # Add Relation to Questionnaire
            value_editor(
                project = project,
                uri = statement['relation'],
                info = {
                    'option': Option.objects.get(uri=props['mapping'][prop]),
                    'collection_index': None,
                    'set_index': index['idx'],
                    'set_prefix': index['set_prefix']
                }
            )

            # Add Relatant to Questionnaire
            value_editor(
                project = project,
                uri = statement['relatant'],
                info = {
                    'text': f"{value.label} ({value.description}) [{source}]",
                    'external_id': value.id,
                    'collection_index': None,
                    'set_index': index['idx'],
                    'set_prefix': index['set_prefix']
                }
            )

            # Add Assumption
            if statement.get('assumption') and hasattr(value, 'qualifier') and value.qualifier:
                # Get Assumptions
                assumption_dict = process_qualifier(value.qualifier)
                # Add Assumptions
                for assumption_key, assumption_value in assumption_dict.items():
                    value_editor(
                        project = project,
                        uri = statement['assumption'],
                        info = {
                            'text': "{label} ({description}) [{source}]".format_map(
                                assumption_value
                            ),
                            'external_id': assumption_value['id'],
                            'collection_index': assumption_key,
                            'set_index': index['idx'],
                            'set_prefix': index['set_prefix']
                        }
                    )

            # Add Order Number
            if statement.get('order') and hasattr(value, 'order') and value.order:
                # Add Order Number to Questionnaire
                value_editor(
                    project = project,
                    uri = statement['order'],
                    info = {
                        'text': value.order,
                        'set_index': index['idx'],
                        'set_prefix': index['set_prefix']
                    }
                )

            # Update index
            index['idx'] += 1

            # Update existing IDs, Texts, and Relations
            info['value_ids'].append(value.id)
            info['set_prefix_ids'].append(index['set_prefix_reduced'])
            info['texts'].append(f"{value.label} ({value.description}) [{source}]")
            info['rels'].append(props['mapping'][prop])

def add_properties(project, data, uri, set_prefix):
    '''Function which adds Data Properties to the Questionnaire.
       
       Input: Data Properties
       Output: -'''

    for key, value in data.properties.items():
        value_editor(
            project = project,
            uri  = uri,
            info = {
                'option': Option.objects.get(uri=value[0]),
                'collection_index': key,
                'set_index': 0,
                'set_prefix': set_prefix
            }
        )

def add_references(project, data, uri, set_index = 0, set_prefix = None):
    '''Function which adds References to the Questionnaire.
       
       Input: References
       Output: -'''

    for key, value in data.reference.items():
        value_editor(
            project = project,
            uri  = uri,
            info = {
                'text': value[1],
                'option': Option.objects.get(uri=value[0]),
                'collection_index': key,
                'set_index': set_index,
                'set_prefix': set_prefix
            }
        )

def add_reference_order(entity):
    '''Function which defines the reference order according to the Item type
    
       Input: Item Information
       Output: -'''
    options = get_options()
    # Define Flexible Key and Value
    flex_key = 'morwiki' if entity == 'benchmark' else 'swmath'
    flex_value = options['MORWIKI' if entity == 'benchmark' else 'SWMATH']
    # Define Reference Order
    order = {
        'doi': (0, options['DOI']),
        flex_key: (1, flex_value),
        'url': (2, options['URL']),
        }
    return order



'''Module containing Utility Functions for the Model Documentation'''

def build_quantity_info(quantity, qtype):
    """Build the Info dict for a Quantity or QuantityKind."""
    base_info = {
        "Type": qtype,
        "Name": quantity.get("Name", ""),
        "Description": quantity.get("Description", ""),
        "ID": (
            quantity.get("ID", "")
            if quantity.get("ID", "") and quantity.get("ID", "") != "not found"
            else quantity.get("Reference", "")
            if qtype == "Quantity" and quantity.get("Reference", "")
            else ""
        ),
    }

    # Only Quantity has QKRelatant
    if qtype == "Quantity":
        base_info["QKName"] = quantity.get("QKRelatant", [{}])[0].get("Name", "")
        base_info["QKID"] = quantity.get("QKRelatant", [{}])[0].get("ID", "")

    return base_info


def map_entity_quantity(data, entity_type, mapping):
    """Map quantities or quantity kinds to entity elements."""
    for entity in data.get(entity_type, {}).values():
        for element in entity.get("element", {}).values():
            element_quantity_name = element.get("quantity", {}).get("Name", "").lower()

            for quantity in data.get("quantity", {}).values():
                quantity_name = quantity.get("Name", "").lower()

                if element_quantity_name != quantity_name:
                    continue

                qtype = quantity.get("QorQK")
                if qtype in mapping.values():
                    element["Info"] = build_quantity_info(quantity, qtype)



'''Module containing Constants for the Publication Documentation'''

PROPS = {
    'P2E': ['documents','invents','studies','surveys','uses'],
    'P2A': ['analyzes','applies','invents','studies','surveys'],
    'P2BS': ['documents', 'uses']
}

# Index counters for different cenchmarks and software
INDEX_COUNTERS = {
    "benchmark": 0,
    "software": 0
}

# URI mappings for quantity relations
RELATION_URIS = {
    "benchmark": "P2B",
    "software": "P2S",
}

# URI mappings for quantity relatants
RELATANT_URIS = {
    "benchmark": "BRelatant",
    "software": "SRelatant",
}

# URI mappings for publications
PUBLICATIONS = {
    "Name": "title",
    "Description": "description",
    "Entrytype": "entrytype",
    "Title": "title",
    "Date": "date",
    "Volume": "volume",
    "Issue": "issue",
    "Page": "page"
}

# URI mappings for languages
LANGUAGES = {
    "Language ID": "id",
    "Language Name": "label",
    "Language Description": "description"
}

# URI mappings for journals
JOURNALS = {
    "Journal ID": "id",
    "Journal ISSN": "issn",
    "Journal Name": "label",
    "Journal Description": "description"
}

# URI mappings for author
AUTHORS = {
    "Author ID": "id",
    "Author ORCID": "orcid_id",
    "Author ZBMath": "zbmath_id",
    "Author Wikidata": "wikidata_id",
    "Author Name": "label",
    "Author Description": "description"
}



from rdmo.options.providers import Provider
from rdmo.domain.models import Attribute

from .sparql import queryProvider

from ..config import BASE_URI, endpoint
from ..getters import get_items, get_data, get_properties, get_questions
from ..helpers import define_setup
from ..queries import query_sources, query_sources_with_user_additions, query_sparql

class MaRDIAndWikidataSearch(Provider):

    search = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        return query_sources(search, query_id, sources)

class MainMathematicalModel(Provider):

    search = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'MM'
        sources = ['mardi']

        return query_sources(search, query_id, sources)
    
class Method(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'AL'
        sources = ['mathalgodb','mardi','wikidata']

        return query_sources(search, query_id, sources)
    
class RelatedMethod(Provider):

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search:
            return []
        
        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['method'],
            query_id = 'AL',
            sources = ['mathalgodb','mardi','wikidata']
        )
        
        return query_sources_with_user_additions(
            search = search, 
            project = project, 
            setup = setup
        )
    
class WorkflowTask(Provider):

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MaRDI Portak for Task related to chosen Model'''

        questions = get_questions('workflow')
        options = []
        model_id = ''

        values = project.values.filter(snapshot=None, attribute=Attribute.objects.get(uri=f'{BASE_URI}{questions["Model"]["ID"]["uri"]}'))        
        for value in values:
            model_id = value.external_id

        if model_id:
            _, id_value = model_id.split(':')
            results = query_sparql(queryProvider['RT'].format(id_value, **get_items(), **get_properties()), endpoint['mardi']['sparql'])
            if results:
                if results[0].get('usedBy', {}).get('value'):
                    tasks = results[0]['usedBy']['value'].split(' / ')
                    for task in tasks:
                        id, label, description = task.split(' | ')
                        options.append({'id': id, 'text': f'{label} ({description}) [mardi]'})

        return options
    
class Software(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = 'SO'
        sources = ['mathalgodb', 'mardi', 'wikidata']

        return query_sources(search, query_id, sources)

class RelatedSoftware(Provider):

    search = True

    def get_options(self, project, search=None, user=None, site=None):
    
        if not search:
            return []
        
        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['software'],
            query_id = 'SO',
            sources = ['mathalgodb','mardi','wikidata']
        )
        
        return query_sources_with_user_additions(
            search = search, 
            project = project, 
            setup = setup
        )

class Hardware(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        return query_sources(search, query_id, sources)
    
class Instrument(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        return query_sources(search, query_id, sources)
    
class DataSet(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        return query_sources(search, query_id, sources)
    
class RelatedInstrument(Provider):

    search = True

    def get_options(self, project, search=None, user=None, site=None):
    
        if not search:
            return []
        
        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['instrument']
        )
        
        return query_sources_with_user_additions(
            search = search, 
            project = project, 
            setup = setup
        )
    
class RelatedDataSet(Provider):

    search = True

    def get_options(self, project, search=None, user=None, site=None):
    
        if not search:
            return []
        
        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['data-set']
        )

        return query_sources_with_user_additions(
            search = search, 
            project = project, 
            setup = setup
        )
            
class ProcessStep(Provider):

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        return query_sources(search, query_id, sources)
    
class Discipline(Provider):

    msc = get_data('data/msc2020.json')

    search = True
    refresh = True

    def get_options(self, project, search, user=None, site=None):
        '''Queries MaRDI Portal and Wikidata for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = ''
        sources = ['mardi', 'wikidata']

        # Discipline from Knowledge Graphs
        options = query_sources(search, query_id, sources, False)

        # Mathematical Subjects
        options.extend([{'id': f"msc:{self.msc[key]['id']}", 'text': f"{key} ({self.msc[key]['quote']}) [msc]"} for key in self.msc if search.lower() in key.lower()])

        return sorted(options, key=lambda option: option['text'].lower())[:30]
